"""
problog.forward - Forward compilation and evaluation
----------------------------------------------------

Forward compilation using TP-operator.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import copy
import logging
import random
import signal
import time
from collections import defaultdict

from .bdd_formula import BDD
from .core import transform
from .core import transform_create_as
from .dd_formula import DD
from .dd_formula import build_dd
from .evaluator import Evaluator, EvaluatableDSP, InconsistentEvidenceError
from .formula import LogicFormula, OrderedSet, atom
from .sdd_formula import SDD
from .util import UHeap


def timeout_handler(signum, frame):
    raise SystemError("Process timeout (Python) [%s]" % signum)


class ForwardInference(DD):
    def __init__(self, compile_timeout=None, **kwdargs):
        super(ForwardInference, self).__init__(auto_compact=False, **kwdargs)

        self._inodes_prev = None
        self._inodes_old = None
        self._inodes_neg = None
        self._facts = None
        self._atoms_in_rules = None
        self._completed = None

        self.timeout = compile_timeout

        self._update_listeners = []

        self._node_depths = None
        self.evidence_node = 0

    def register_update_listener(self, obj):
        self._update_listeners.append(obj)

    def _create_atom(
        self, identifier, probability, group, name=None, source=None, is_extra=False
    ):
        return atom(identifier, probability, group, name, source, is_extra)

    def is_complete(self, node):
        node = abs(node)
        return self._completed[node - 1]

    def set_complete(self, node):
        self._completed[node - 1] = True

    def init_build(self):
        if self.evidence():
            ev = [n for q, n in self.evidence() if n is None or n != 0]
            if ev:
                if len(ev) == 1:
                    self.evidence_node = ev[0]
                else:
                    self.evidence_node = self.add_and(ev)
            else:
                # Only deterministically true evidence
                self.evidence_node = 0

        self._facts = []  # list of facts
        self._atoms_in_rules = defaultdict(
            OrderedSet
        )  # lookup all rules in which an atom is used
        self._completed = [False] * len(self)

        self._compute_node_depths()
        for index, node, nodetype in self:
            if self._node_depths[index - 1] is not None:
                # only include nodes that are reachable from a query or evidence
                if nodetype == "atom":  # it's a fact
                    self._facts.append(index)
                    self.set_complete(index)
                else:  # it's a compound
                    for atom in node.children:
                        self._atoms_in_rules[abs(atom)].add(index)
        self.build_constraint_dd()
        self.inodes = [None] * len(self)
        self._inodes_prev = [None] * len(self)
        self._inodes_old = [None] * len(self)
        self._inodes_neg = [None] * len(self)
        self._compute_minmax_depths()

    def _propagate_complete(self, interrupted=False):
        if not interrupted:
            for i, c in enumerate(self._completed):
                if not c:
                    self._completed[i] = True
                    self.notify_node_completed(i + 1)
        else:
            updated_nodes = set([(i + 1) for i, c in enumerate(self._completed) if c])
            while updated_nodes:
                next_updates = set()
                # Find all heads that are affected
                affected_nodes = set()
                for node in updated_nodes:
                    for rule in self._atoms_in_rules[node]:
                        if not self.is_complete(rule):
                            affected_nodes.add(rule)
                for head in affected_nodes:
                    # head must be compound
                    node = self.get_node(head)
                    children = [self.is_complete(c) for c in node.children]
                    if False not in children:
                        self.is_complete(head)
                        self.notify_node_completed(head)
                        next_updates.add(head)
                updated_nodes = next_updates

    def _compute_node_depths(self):
        """Compute node depths in breadth-first manner."""
        self._node_depths = [None] * len(self)
        self._node_levels = []
        # Start with current nodes
        current_nodes = set(
            abs(n) for q, n, l in self.labeled() if self.is_probabilistic(n)
        )
        if self.is_probabilistic(self.evidence_node):
            current_nodes.add(abs(self.evidence_node))
        current_level = 0
        while current_nodes:
            self._node_levels.append(current_nodes)
            next_nodes = set()
            for index in current_nodes:
                self._node_depths[index - 1] = current_level
                node = self.get_node(index)
                nodetype = type(node).__name__
                if nodetype != "atom":
                    for c in node.children:
                        if self.is_probabilistic(c):
                            if self._node_depths[abs(c) - 1] is None:
                                next_nodes.add(abs(c))
            current_nodes = next_nodes
            current_level += 1

    def _compute_minmax_depths(self):
        self._node_minmax = [None] * len(self)
        for level, nodes in reversed(list(enumerate(self._node_levels))):
            for index in nodes:
                # Get current node's minmax
                minmax = self._node_minmax[index - 1]
                if minmax is None:
                    minmax = level
                for rule in self._atoms_in_rules[index]:
                    rule_minmax = self._node_minmax[rule - 1]
                    if rule_minmax is None:
                        self._node_minmax[rule - 1] = minmax
                    else:
                        node = self.get_node(rule)
                        nodetype = type(node).__name__
                        if nodetype == "conj":
                            rule_minmax = max(minmax, rule_minmax)
                        else:  # disj
                            rule_minmax = min(minmax, rule_minmax)
                        self._node_minmax[rule - 1] = rule_minmax

    def _update_minmax_depths(self, index, new_minmax=0):
        """Update the minmax depth data structure when the given node is completed.

        :param index:
        :return:
        """
        current_minmax = self._node_minmax[index - 1]
        self._node_minmax[index - 1] = new_minmax

        for parent in self._atoms_in_rules[index]:
            parent_minmax = self._node_minmax[parent - 1]

            if current_minmax == parent_minmax:
                # Current node is best child => we need to recompute
                parent_node = self.get_node(parent)
                parent_nodetype = type(parent_node).__name__
                parent_children_minmax = [
                    self._node_minmax[c - 1]
                    for c in parent_node.children
                    if not self.is_complete(c)
                ]
                if not parent_children_minmax:
                    # No incomplete children
                    self.set_complete(parent)
                    parent_minmax = 0
                elif parent_nodetype == "conj":
                    parent_minmax = max(parent_children_minmax)
                else:
                    parent_minmax = min(parent_children_minmax)
                self._update_minmax_depths(parent, parent_minmax)

    def sort_nodes(self, nodes):
        return sorted(nodes, key=lambda i: self._node_depths[i - 1])

    def notify_node_updated(self, node, complete):
        for obj in self._update_listeners:
            obj.node_updated(self, node, complete)

    def notify_node_completed(self, node):
        for obj in self._update_listeners:
            obj.node_completed(self, node)

    def _heuristic_key_depth(self, node):
        # For OR: D(n) is min(D(c) for c in children)
        # For AND: D(n) is max(D(c) for c in children)
        return self._node_minmax[node - 1], self._node_depths[node - 1], random.random()

    def _heuristic_key(self, node):
        return self._heuristic_key_depth(node)

    def build_iteration(self, updated_nodes):
        to_recompute = UHeap(key=self._heuristic_key)
        for node in updated_nodes:
            for rule in self._atoms_in_rules[node]:
                to_recompute.push(rule)

        # nodes_to_recompute should be an updateable heap without duplicates
        while to_recompute:
            key, node = to_recompute.pop_with_key()
            if self.update_inode(node):  # The node has changed
                # Find rules that may be affected
                for rule in self._atoms_in_rules[node]:
                    to_recompute.push(rule)
                # Notify listeners that node was updated
                self.notify_node_updated(node, self.is_complete(node))
            elif self.is_complete(node):
                self.notify_node_completed(node)
                # if self.is_complete(node):
                #     self._update_minmax_depths(node)

    def build_iteration_levelwise(self, updated_nodes):
        while updated_nodes:
            next_updates = OrderedSet()
            # Find all heads that are affected
            affected_nodes = OrderedSet()
            for node in updated_nodes:
                for rule in self._atoms_in_rules[node]:
                    affected_nodes.add(rule)
            affected_nodes = self.sort_nodes(affected_nodes)
            # print (affected_nodes, [self._node_depths[i-1] for i in affected_nodes])
            for head in affected_nodes:
                if self.update_inode(head):
                    next_updates.add(head)
                    self.notify_node_updated(head, self.is_complete(head))
                elif self.is_complete(head):
                    self.notify_node_completed(head)
            updated_nodes = next_updates

    def build_stratum(self, updated_nodes):
        self.build_iteration(updated_nodes)
        updated_nodes = OrderedSet()
        for i, nodes in enumerate(zip(self.inodes, self._inodes_old)):
            if not self.get_manager().same(*nodes):
                updated_nodes.add(i + 1)
                # self.notify_node_updated(i + 1)
        self.get_manager().ref(*filter(None, self.inodes))
        self.get_manager().deref(*filter(None, self._inodes_prev))
        self.get_manager().deref(*filter(None, self._inodes_neg))

        # Only completed nodes should be used for negation in the next stratum.
        self._inodes_old = self.inodes[:]
        self._inodes_prev = [None] * len(self)
        for i, n in enumerate(self.inodes):
            if self._completed[i]:
                self._inodes_prev[i] = n
        self._inodes_neg = [None] * len(self)

        return updated_nodes

    def build_dd(self):
        required_nodes = set(
            [abs(n) for q, n, l in self.labeled() if self.is_probabilistic(n)]
        )
        required_nodes |= set(
            [abs(n) for q, n, v in self.evidence_all() if self.is_probabilistic(n)]
        )
        if self.timeout:
            # signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(self.timeout)
            signal.signal(signal.SIGALRM, timeout_handler)
            logging.getLogger("problog").info("Set timeout:", self.timeout)
        try:
            self.init_build()
            updated_nodes = OrderedSet(self._facts)
            while updated_nodes:
                # TODO only check nodes that are actually used in negation
                updated_nodes = self.build_stratum(updated_nodes)
            self._propagate_complete(False)
        except SystemError as err:
            self._propagate_complete(True)
            logging.getLogger("problog").warning(err)
        except KeyboardInterrupt as err:
            self._propagate_complete(True)
            logging.getLogger("problog").warning(err)

        signal.alarm(0)
        self.build_constraint_dd()

    def current(self):
        destination = LogicFormula(auto_compact=False)
        source = self
        # TODO maintain a translation table
        for i, n, t in source:
            inode = self.get_inode(i)
            if inode is not None:
                inode = int(inode)
            if t == "atom":
                j = destination.add_atom(
                    n.identifier,
                    n.probability,
                    n.group,
                    name=inode,
                    is_extra=n.is_extra,
                )
            elif t == "conj":
                children = [c for c in n.children if self.get_inode(c) is not None]
                j = destination.add_and(children, name=inode)
            elif t == "disj":
                children = [c for c in n.children if self.get_inode(c) is not None]
                j = destination.add_or(children, name=inode)
            else:
                raise TypeError("Unknown node type")
            assert i == j

        for name, node, label in source.get_names_with_label():
            if label != self.LABEL_NAMED:
                destination.add_name(name, node, label)

        for c in source.constraints():
            if c.is_nontrivial():
                destination.add_constraint(c)
        return destination

    def update_inode(self, index):
        """Recompute the inode at the given index."""

        was_complete = self.is_complete(index)

        oldnode = self.get_inode(index)
        node = self.get_node(index)
        assert index > 0
        nodetype = type(node).__name__
        if nodetype == "conj":
            children = [self.get_inode(c) for c in node.children]
            children_complete = [self.is_complete(c) for c in node.children]
            if None in children:
                newnode = None  # don't compute if some children are still unknown
            else:
                newnode = self.get_manager().conjoin(*children)
            if False not in children_complete:
                self.set_complete(index)

        elif nodetype == "disj":
            children = [self.get_inode(c) for c in node.children]
            children_complete = [self.is_complete(c) for c in node.children]
            children = list(
                filter(None, children)
            )  # discard children that are still unknown
            if children:
                newnode = self.get_manager().disjoin(*children)
            else:
                newnode = None
            if False not in children_complete:
                self.set_complete(index)

        else:
            raise TypeError("Unexpected node type.")

        # Add constraints
        if newnode is not None:
            newernode = self.get_manager().conjoin(newnode, self.get_constraint_inode())
            self.get_manager().deref(newnode)
            newnode = newernode

        if self.get_manager().same(oldnode, newnode):
            return self.is_complete(index) != was_complete  # no change occurred
        else:
            if oldnode is not None:
                self.get_manager().deref(oldnode)
            self.set_inode(index, newnode)
            return True

    def get_evidence_inode(self):
        if not self.is_probabilistic(self.evidence_node):
            return self.get_manager().true()
        else:
            inode = self.get_inode(self.evidence_node, final=True)
            if inode:
                return inode
            else:
                return self.get_manager().true()

    def get_inode(self, index, final=False):
        """
        Get the internal node corresponding to the entry at the given index.
        :param index: index of node to retrieve
        :return: SDD node corresponding to the given index
        :rtype: SDDNode
        """
        assert self.is_probabilistic(index)
        node = self.get_node(abs(index))
        if type(node).__name__ == "atom":
            av = self.atom2var.get(abs(index))
            if av is None:
                av = self.get_manager().add_variable()
                self.atom2var[abs(index)] = av
                self.var2atom[av] = abs(index)
            result = self.get_manager().literal(av)
            if index < 0:
                return self.get_manager().negate(result)
            else:
                return result
        elif index < 0 and not final:
            # We are requesting a negated node => use previous stratum's result
            result = self._inodes_neg[-index - 1]
            if result is None and self._inodes_prev[-index - 1] is not None:
                result = self.get_manager().negate(self._inodes_prev[-index - 1])
                self._inodes_neg[-index - 1] = result
            return result
        elif index < 0:
            return self.get_manager().negate(self.inodes[-index - 1])
        else:
            return self.inodes[index - 1]

    def set_inode(self, index, node):
        """Set the internal node for the given index.

        :param index: index at which to set the new node
        :type index: int > 0
        :param node: new node
        """
        assert index is not None
        assert index > 0
        self.inodes[index - 1] = node

    def add_constraint(self, c):
        LogicFormula.add_constraint(self, c)


class _ForwardSDD(SDD, ForwardInference):
    transform_preference = 1000

    def __init__(self, sdd_auto_gc=False, **kwdargs):
        SDD.__init__(self, sdd_auto_gc=sdd_auto_gc, **kwdargs)
        ForwardInference.__init__(self, **kwdargs)

    @classmethod
    def is_available(cls):
        return SDD.is_available()

    def to_explicit_encoding(self):
        """
        Transform the current implicit encoding to an SDD with explicit encoding. The latter will contain indicator
        variables for inferred literals and contains both the 'true and false' case. For example, while the implicit
        encoding for c :- a,b will result in c = a * b., the explicit encoding will result in (c * a * b) + (-c * (-a + -b))
        In the explicit encoding, evidence can be incorporated as changing the weights and a negation can also easily be
        queried. Furthermore, the explicit encoding implementation can guarantee a circuit with one root instead of
        multiple.

        :return: The explicit encoding of this formula.
        :rtype: SDDExplicit
        """
        from .sdd_formula_explicit import build_explicit_from_forwardsdd
        from .sdd_formula_explicit import SDDExplicit

        return build_explicit_from_forwardsdd(source=self, destination=SDDExplicit())

    def copy_to_noref(self, destination):
        """
        Copy the relevant data structures without any refcounts to the new inodes.
        Because of this, the destination should have auto_gc disabled.
        :param destination: The DD to copy the data structures' references to.
        :type destination: DD
        """
        destination.atom2var = self.atom2var.copy()
        destination.var2atom = self.var2atom.copy()

        old = self.inode_manager.nodes
        self.inode_manager.nodes = self.inodes
        destination.inode_manager = self.inode_manager.get_deepcopy_noref()
        self.inode_manager.nodes = old

        destination._atomcount = self._atomcount
        destination._weights = self._weights.copy()
        destination._constraints = self._constraints.copy()
        destination.evidence_node = self.evidence_node
        destination._nodes = self._nodes.copy()
        destination._index_atom = self._index_atom.copy()
        destination._index_next = self._index_next
        destination._index_conj = self._index_conj.copy()
        destination._index_disj = self._index_disj.copy()
        destination._names = copy.deepcopy(self._names)
        destination._constraints_me = copy.deepcopy(self._constraints_me)


class _ForwardBDD(BDD, ForwardInference):
    transform_preference = 1000

    def __init__(self, **kwdargs):
        BDD.__init__(self, **kwdargs)
        ForwardInference.__init__(self, **kwdargs)

    @classmethod
    def is_available(cls):
        return BDD.is_available()


@transform(LogicFormula, _ForwardSDD)
def build_sdd(source, destination, **kwdargs):
    result = build_dd(source, destination, **kwdargs)
    return result


@transform(LogicFormula, _ForwardBDD)
def build_sdd(source, destination, **kwdargs):
    result = build_dd(source, destination, **kwdargs)
    return result


class ForwardSDD(LogicFormula, EvaluatableDSP):
    transform_preference = 30

    def __init__(self, **kwargs):
        LogicFormula.__init__(self, **kwargs)
        EvaluatableDSP.__init__(self)
        self.kwargs = kwargs
        self.internal = _ForwardSDD(**self.kwargs)

    @classmethod
    def is_available(cls):
        return SDD.is_available()

    def _create_evaluator(self, semiring, weights, **kwargs):
        return ForwardEvaluator(self, semiring, self.internal, weights, **kwargs)

    def to_formula(self):
        build_dd(self, self.internal)
        return self.internal.to_formula()


class ForwardBDD(LogicFormula, EvaluatableDSP):
    transform_preference = 40

    def __init__(self, **kwargs):
        LogicFormula.__init__(self, **kwargs)
        EvaluatableDSP.__init__(self)
        self.kwargs = kwargs

    @classmethod
    def is_available(cls):
        return BDD.is_available()

    def _create_evaluator(self, semiring, weights, **kwargs):
        return ForwardEvaluator(
            self, semiring, _ForwardBDD(**self.kwargs), weights, **kwargs
        )


# Inform the system that we can create a ForwardFormula in the same way as a LogicFormula.
transform_create_as(ForwardSDD, LogicFormula)
transform_create_as(ForwardBDD, LogicFormula)


class ForwardEvaluator(Evaluator):
    """An evaluator using anytime forward compilation."""

    def __init__(self, formula, semiring, fdd, weights=None, verbose=None, **kwargs):
        Evaluator.__init__(self, formula, semiring, weights, **kwargs)

        self.fsdd = fdd
        self._z = None
        self._verbose = verbose

        self._results = {}
        self._complete = set()

        self._start_time = None

    def node_updated(self, source, node, complete):
        is_evidence = node == abs(source.evidence_node)
        name = [
            n
            for n, i, l in self.formula.labeled()
            if source.is_probabilistic(i) and abs(i) == node
        ]
        if is_evidence:
            name = ("evidence",)
        if name:
            name = name[0]
            weights = {}
            for atom, weight in self.weights.items():
                av = source.atom2var.get(atom)
                if av is not None:
                    weights[av] = weight
            inode = source.get_inode(node, final=True)
            if inode is not None:
                enode = source.get_manager().conjoin(
                    source.get_evidence_inode(), source.get_constraint_inode()
                )
                if self.fsdd.get_manager().is_false(enode):
                    raise InconsistentEvidenceError(context=" during compilation")

                qnode = source.get_manager().conjoin(inode, enode)
                tvalue = source.get_manager().wmc(enode, weights, self.semiring)
                value = source.get_manager().wmc(qnode, weights, self.semiring)
                result = self.semiring.normalize(value, tvalue)
                self._results[node] = result

                debug_msg = "update query %s: %s after %ss" % (
                    name,
                    self.semiring.result(result, self.formula),
                    "%.4f" % (time.time() - self._start_time),
                )
                logging.getLogger("problog").debug(debug_msg)

            if complete:
                self._complete.add(node)
        # if is_evidence:
        #     for c in self._complete:
        #         if c != node:
        #             self.node_updated(source, c, complete)

    def node_completed(self, source, node):
        qs = set(
            abs(qi) for qn, qi, ql in source.labeled() if source.is_probabilistic(qi)
        )
        if node in qs:
            self._complete.add(node)

    def initialize(self):
        self.weights = self.formula.extract_weights(self.semiring, self.given_weights)

        # We should do all compilation here.
        self.fsdd.register_update_listener(self)
        self._start_time = time.time()
        if len(self.fsdd) == 0:
            if self.fsdd.init_varcount == -1:
                self.fsdd.init_varcount = self.formula.atomcount
            build_dd(self.formula, self.fsdd)

        # Update weights with constraints and evidence
        enode = self.fsdd.get_manager().conjoin(
            self.fsdd.get_evidence_inode(), self.fsdd.get_constraint_inode()
        )

        if self.fsdd.get_manager().is_false(enode):
            raise InconsistentEvidenceError(context=" during compilation")

        # Make sure all atoms exist in atom2var.
        for name, node, label in self.fsdd.labeled():
            if self.fsdd.is_probabilistic(node):
                self.fsdd.get_inode(node)

        weights = {}
        for atom, weight in self.weights.items():
            av = self.fsdd.atom2var.get(atom)
            if av is not None:
                weights[av] = weight
            elif atom == 0:
                weights[0] = weight

        for name, node, label in self.fsdd.labeled():
            if self.fsdd.is_probabilistic(node):
                inode = self.fsdd.get_inode(node)
                qnode = self.fsdd.get_manager().conjoin(inode, enode)
                tvalue = self.fsdd.get_manager().wmc(enode, weights, self.semiring)
                value = self.fsdd.get_manager().wmc(qnode, weights, self.semiring)
                result = self.semiring.normalize(value, tvalue)
            elif self.fsdd.is_true(node):
                result = self.semiring.one()
            else:
                result = self.semiring.zero()
            self._results[node] = result

    def propagate(self):
        self.initialize()

    def evaluate(self, index):
        """Compute the value of the given node."""
        # We should get results from cache here.

        ub = 1.0
        if index is None:
            return 0.0
        elif index == 0:
            if not self.semiring.is_nsp():
                return self.semiring.result(self.semiring.one(), self.formula)
            else:
                weights = {}
                for atom, weight in self.weights.items():
                    av = self.fsdd.atom2var.get(atom)
                    if av is not None:
                        weights[av] = weight
                    elif atom == 0:
                        weights[0] = weight
                enode = self.fsdd.get_manager().conjoin(
                    self.fsdd.get_evidence_inode(), self.fsdd.get_constraint_inode()
                )
                tvalue = self.fsdd.get_manager().wmc(enode, weights, self.semiring)
                result = self.semiring.normalize(tvalue, tvalue)
                return self.semiring.result(result, self.formula)
        else:
            n = self.formula.get_node(abs(index))
            nt = type(n).__name__
            if nt == "atom":
                wp = self._results[index]
                # wp, wn = self.weights.get(abs(index))
                if index < 0:
                    # wn = self.semiring.negate(wp)
                    return self.semiring.result(wp, self.formula)
                else:
                    return self.semiring.result(wp, self.formula)
            else:
                # TODO report correct bounds in case of evidence
                if index < 0:
                    if -index in self._results:
                        if -index in self._complete:
                            return self.semiring.result(
                                self.semiring.negate(self._results[-index]),
                                self.formula,
                            )
                        else:
                            return (
                                self.semiring.result(self.semiring.zero()),
                                self.semiring.result(
                                    self.semiring.negate(self._results[-index]),
                                    self.formula,
                                ),
                            )
                    else:
                        return (
                            self.semiring.result(self.semiring.zero()),
                            self.semiring.result(self.semiring.one()),
                        )
                else:
                    if index in self._results:
                        if index in self._complete:
                            return self.semiring.result(
                                self._results[index], self.formula
                            )
                        else:
                            return (
                                self.semiring.result(
                                    self._results[index], self.formula
                                ),
                                self.semiring.result(self.semiring.one()),
                            )
                    else:
                        return (
                            self.semiring.result(self.semiring.zero()),
                            self.semiring.result(self.semiring.one()),
                        )

    def evaluate_evidence(self):
        raise NotImplementedError("Evaluator.evaluate_evidence is an abstract method.")

    # def add_evidence(self, node):
    #     """Add evidence"""
    #     warnings.warn('Evidence is not supported by this evaluation method and will be ignored.')

    def has_evidence(self):
        return self.__evidence != []

    def clear_evidence(self):
        self.__evidence = []

    def evidence(self):
        return iter(self.__evidence)



"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), "..", ".."))

from problog.util import load_module
from problog import version

problog_tasks = {}
problog_tasks["prob"] = "problog.tasks.probability"
problog_tasks["mpe"] = "problog.tasks.mpe"
problog_tasks["sample"] = "problog.tasks.sample"
problog_tasks["ground"] = "problog.tasks.ground"
problog_tasks["lfi"] = "problog.learning.lfi"
problog_tasks["explain"] = "problog.tasks.explain"
problog_tasks["web"] = "problog.web.server"
problog_tasks["dt"] = "problog.tasks.dtproblog"
problog_tasks["shell"] = "problog.tasks.shell"
problog_tasks["bn"] = "problog.tasks.bayesnet"
problog_tasks["parse"] = "problog.parser"
problog_tasks["map"] = "problog.tasks.map"
problog_tasks["time"] = "problog.tasks.time1"
problog_tasks["constraint"] = "problog.tasks.constraint"

problog_default_task = "prob"


def run_task(argv):
    """Execute a task in ProbLog.
    If the first argument is a known task name, that task is executed.
    Otherwise the default task is executed.

    :param argv: list of arguments for the task
    :return: result of the task (typically None)
    """
    if len(argv) > 0 and argv[0] in problog_tasks:
        task = argv[0]
        args = argv[1:]
    else:
        task = problog_default_task
        args = argv
    return load_task(task).main(args)


def load_task(name):
    """Load the module for executing the given task.

    :param name: task name
    :type name: str
    :return: loaded module
    :rtype: module
    """
    return load_module(problog_tasks[name])


def main(argv=None):
    """Main function.
    :param argv: command line arguments
    """
    if argv is None:
        argv = sys.argv[1:]

    if len(argv) > 0:
        if argv[0] == "install":
            from .. import setup

            setup.install()
            return
        elif argv[0] == "info":
            from problog.core import list_transformations

            list_transformations()
            return
        elif argv[0] == "unittest":
            import unittest

            test_results = unittest.TextTestResult(sys.stderr, False, 1)
            unittest.TestLoader().discover(
                os.path.join(os.path.dirname(__file__), "..")
            ).run(test_results)
            return
        elif argv[0] == "--version":
            print(version.version)
            return
        else:
            return run_task(argv)
    else:
        return run_task(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



"""

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import argparse
import json
import sys
import traceback

from ..engine import DefaultEngine
from ..errors import process_error
from ..kbest import KBestFormula
from ..program import PrologFile
from ..util import init_logger, format_dictionary


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("filename")
    parser.add_argument("-v", "--verbose", action="count")
    parser.add_argument("--web", action="store_true")
    parser.add_argument("-o", "--output", type=str, default=None)
    args = parser.parse_args(argv)

    init_logger(args.verbose)
    if args.output:
        out = open(args.output, "w")
    else:
        out = sys.stdout

    result = {}
    try:
        pl = PrologFile(args.filename)
        db = DefaultEngine().prepare(pl)

        program = list(map(lambda s: "%s." % s, db.iter_raw()))
        cnf = KBestFormula.create_from(db, label_all=True)

        explanation = []
        results = cnf.evaluate(explain=explanation)

        result["SUCCESS"] = True
        result["program"] = program
        result["proofs"] = explanation
        result["results"] = results
        if args.web:
            result["probabilities"] = [
                (str(k), round(v, 8)) for k, v in results.items()
            ]
            print(json.dumps(result), file=out)
        else:
            print("Transformed program", file=out)
            print("-------------------", file=out)
            print("\n".join(program), file=out)
            print(file=out)

            print("Proofs", file=out)
            print("------", file=out)
            print("\n".join(explanation), file=out)

            print(file=out)
            print("Probabilities", file=out)
            print("-------------", file=out)
            print(format_dictionary(results), file=out)

    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        if args.web:
            result["SUCCESS"] = False
            result["err"] = vars(err)
            result["err"]["message"] = process_error(err)
            print(json.dumps(result), file=out)
        else:
            print(process_error(err), file=out)

    if args.output:
        out.close()

    return result


if __name__ == "__main__":
    main(sys.argv[1:])


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("inputfile")
    return parser


if __name__ == "__main__":
    main(**vars(argparser().parse_args()))



import os
from collections import defaultdict, deque
from subprocess import CalledProcessError

from .engine import UnknownClause
from .errors import GroundingError
from .logic import AnnotatedDisjunction, list2term, Term, Clause
from .util import subprocess_check_output, mktempfile, Timer


def ground_yap(
    model,
    target=None,
    queries=None,
    evidence=None,
    propagate_evidence=False,
    labels=None,
    engine=None,
    debug=False,
    **kwdargs
):
    """Ground a given model.

    :param model: logic program to ground
    :type model: LogicProgram
    :param target: formula in which to store ground program
    :type target: LogicFormula
    :param queries: list of queries to override the default
    :param evidence: list of evidence atoms to override the default
    :return: the ground program
    :rtype: LogicFormula
    """

    with Timer("Grounding (YAP)"):

        if debug:
            fn_model = "/tmp/model.pl"
            fn_ground = "/tmp/model.ground"
            fn_evidence = "/tmp/model.evidence"
            fn_query = "/tmp/model.query"

        else:
            fn_model = mktempfile(".pl")
            fn_ground = mktempfile(".ground")
            fn_evidence = mktempfile(".evidence")
            fn_query = mktempfile(".query")

        with open(fn_model, "w") as f:
            f.write("\n".join(map(statement_to_yap, model)) + "\n")

        yap_ground = os.path.join(os.path.dirname(__file__), "yap", "ground_compact.pl")

        cmd = [
            "yap",
            "-L",
            yap_ground,
            "--",
            fn_model,
            fn_ground,
            fn_evidence,
            fn_query,
        ]

        try:
            output = subprocess_check_output(cmd)

        except CalledProcessError as err:
            errmsg = err.output.strip()
            if errmsg.startswith("undefined"):
                raise UnknownClause(errmsg.split()[1], None)
            elif errmsg.startswith("non-ground"):
                raise GroundingError(
                    "Non-ground clause detected: %s" % errmsg.split()[1], None
                )
            else:
                raise err

        with open(fn_query) as f:
            queries = f.readlines()

        with open(fn_evidence) as f:
            evidence = f.readlines()

        with open(fn_ground) as f:
            return read_grounding(f, target, queries, evidence)


def read_grounding(lines, target, queries, evidence):
    """

    :param lines:
    :param target:
    :type target: LogicFormula
    :return:
    """

    id_counts = defaultdict(int)
    parsed_lines = deque()
    names = {}

    # Iteration 1: read in the file and determine which line numbers occur multiple times
    line_nums = []
    for i, line in enumerate(lines):
        line = line.strip()
        line, name = line.split("|")
        name = name.strip()
        name = Term.from_string(name)

        line_num, line_type, line_content = line.split(None, 2)
        line_num = int(line_num)
        names[name] = line_num

        # Count how many times each line occurs
        id_counts[line_num] += 1

        line_nums.append(line_num)
        if line_type == "FACT":
            parsed_lines.append((line_num, "FACT", [i, float(line_content)], name))
        elif line_type == "AND":
            parsed_lines.append(
                (line_num, "AND", [int(x) for x in line_content.split()], name)
            )
        else:
            raise Exception("Unexpected type: %s" % line_type)

    num2index = {}

    while parsed_lines:
        line_num, line_type, line_content, line_name = parsed_lines.popleft()

        if line_num in num2index:
            # There is already a node for this line number => add this as a disjunct
            or_node = num2index[line_num]
        elif id_counts[line_num] > 1:
            # It's an or-node
            or_node = target.add_or((), placeholder=True)
            num2index[line_num] = or_node
        else:
            or_node = None

        if line_type == "FACT":
            node_id = target.add_atom(
                line_content[0], probability=line_content[1], name=line_name
            )
        else:  # AND
            found_all = True
            children = []
            for child in line_content:
                negated = child < 0
                child = abs(child)
                child_key = num2index.get(child)
                if child_key is None:
                    if id_counts[child] > 1:
                        # Child is and or-node that doesn't exist yet => we can make a placeholder
                        child_key = target.add_or((), placeholder=True)
                        num2index[child] = child_key
                    else:
                        found_all = False
                        child_key = None
                        break
                if child_key is not None:
                    if negated:
                        children.append(-child_key)
                    else:
                        children.append(child_key)
            if found_all:
                node_id = target.add_and(children, name=line_name)
            else:
                parsed_lines.append((line_num, line_type, line_content, line_name))
                node_id = None

        if node_id is not None:
            if or_node is None:
                num2index[line_num] = node_id
            else:
                target.add_disjunct(or_node, node_id)

    for q in queries:
        q = Term.from_string(q)
        if q not in names:
            key = None
        else:
            key = num2index[names[q]]
        target.add_query(q, key)

    for q in evidence:
        q, v = q.rsplit(None, 1)
        q = Term.from_string(q)
        if q not in names:
            key = None
        else:
            key = num2index[names[q]]
        target.add_evidence(q, key, v == "t")

    return target


def statement_to_yap(statement):
    if isinstance(statement, Clause) and statement.head.functor == "_directive":
        if statement.body.functor in ("consult", "use_module"):
            return ""
        else:
            return ":- %s." % statement.body

    if isinstance(statement, AnnotatedDisjunction):
        heads = statement.heads

        # heads = [Term('', h.with_probability]

        head = Term("problog_ad", list2term(heads), statement.body)

        return "%s." % head
    else:
        return "%s." % statement



"""
Module name
"""

from __future__ import print_function

from problog import root_path
from problog.util import subprocess_call, subprocess_check_output
import unittest
import os
import sys
import glob

if __name__ == "__main__":
    sys.path.insert(
        0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    )


class TestLFICont(unittest.TestCase):
    def setUp(self):
        try:
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError:
            self.assertSequenceEqual = self.assertCountEqual


def createTestLFICont(filename):
    def test(self):
        problogcli = root_path("problog-cli.py")

        model = filename
        examples = filename.replace(".pl", ".ev")
        if not os.path.exists(examples):
            raise Exception("Evidence file is missing: {}".format(examples))

        out = subprocess_check_output(
            [
                sys.executable,
                problogcli,
                "lfi",
                "-n",
                "10",
                "-O",
                model.replace(".pl", ".l_pl"),
                model,
                examples,
            ]
        )
        outline = out.strip().split()
        print(outline)

    return test


# if __name__ == "__main__":
#     filenames = sys.argv[1:]
# else:
#     filenames = glob.glob(root_path("test", "lficont", "*.pl"))
#
#
# for testfile in filenames:
#     testname = "test_lficont_" + os.path.splitext(os.path.basename(testfile))[0]
#     setattr(TestLFICont, testname, createTestLFICont(testfile))
#
#
# if __name__ == "__main__":
#     suite = unittest.TestLoader().loadTestsFromTestCase(TestLFICont)
#     unittest.TextTestRunner(verbosity=2).run(suite)



"""
problog.sdd_formula - Sentential Decision Diagrams
--------------------------------------------------

Interface to Sentential Decision Diagrams (SDD)

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from collections import namedtuple

from .core import transform
from .dd_formula import DD, build_dd, DDManager, DDEvaluator
from .errors import InstallError, InconsistentEvidenceError
from .evaluator import SemiringLogProbability, SemiringProbability
from .formula import LogicDAG, LogicFormula
from .util import mktempfile

# noinspection PyBroadException
try:
    from pysdd import sdd
    from pysdd.iterator import SddIterator
    from pysdd.util import sdd_to_dot
    from pysdd.sdd import Vtree
except Exception as err:
    sdd = None


class SDD(DD):
    """A propositional logic formula consisting of and, or, not and atoms represented as an SDD.

    This class has two restrictions with respect to the default LogicFormula:

        * The number of atoms in the SDD should be known at construction time.
        * It does not support updatable nodes.

    This means that this class can not be used directly during grounding.
    It can be used as a target for the ``makeAcyclic`` method.
    """

    transform_preference = 10

    def __init__(
        self, sdd_auto_gc=False, var_constraint=None, init_varcount=-1, **kwdargs
    ):
        """
        Create an SDD

        :param sdd_auto_gc: Auto garbage collection and minimize (during disjoin and conjoin)
        :param var_constraint: A variable ordering constraint. Currently only x_constrained namedtuple are allowed.
        :type var_constraint: x_constrained
        :param init_varcount: The amount of variables to initialize the manager with.
        :param kwdargs:
        :raise InstallError: When the SDD library is not available.
        """
        if sdd is None:
            raise InstallError(
                "The SDD library is not available. Please install the PySDD package."
            )
        self.auto_gc = sdd_auto_gc
        self._var_constraint = var_constraint
        self._init_varcount = init_varcount
        DD.__init__(self, auto_compact=False, **kwdargs)

    @property
    def var_constraint(self):
        return self._var_constraint

    @var_constraint.setter
    def var_constraint(self, var_constraint):
        """
        Set the variable constraint
        :param var_constraint: A variable ordering constraint. Currently only x_constrained namedtuple are allowed.
        :type var_constraint: x_constrained
        """
        self._var_constraint = var_constraint

    @property
    def init_varcount(self):
        return self._init_varcount

    @init_varcount.setter
    def init_varcount(self, value=0):
        """Set the varcount with which to initialise the manager. Only call before calling the manager """
        assert self.inode_manager is None
        self._init_varcount = value

    def _create_manager(self):
        return SDDManager(
            auto_gc=self.auto_gc,
            var_constraint=self.var_constraint,
            varcount=self.init_varcount,
        )

    def _create_evaluator(self, semiring, weights, **kwargs):
        return SDDEvaluator(self, semiring, weights, **kwargs)

    @classmethod
    def is_available(cls):
        """Checks whether the SDD library is available."""
        return sdd is not None

    def to_internal_dot(self, node=None):
        """
        SDD for the given node, formatted for use with Graphviz dot.

        :param node: The node to get the dot from.
        :type node: SddNode
        :return: The dot format of the given node. When node is None, the shared_sdd will be used (contains all active
        sdd structures).
        :rtype: str
        """
        return self.get_manager().to_internal_dot(node=node)

    def sdd_to_dot(self, node, litnamemap=None, show_id=False, merge_leafs=False):
        """
        SDD for the given node, formatted for use with Graphviz dot. This method provides more control over the used
        symbols than to_internal_dot (see litnamemap). Primes are given by a dotted line, subs by a full line.

        :param node: The node to get the dot from.
        :type node: SddNode
        :param litnamemap: A dictionary providing the symbols to use. The following options are available:
            1. literals, e.g. {1:'A', -1:'-A', ...},
            2. True/False, e.g. {true':'1', 'false':'0'}
            3. And/Or e.g. {'mult':'x', 'add':'+'}
            When litnamemap = True, self.get_litnamemap() will be used.
        :type litnamemap: dict[(int | str), str] | bool | None
        :param show_id: Whether to display the ids of each sdd node.
        :param merge_leafs: Whether to merge the same leaf nodes. True results in less nodes but makes it harder to
        render without having crossing lines.
        :return: The dot format of the given node. When node is None, this mgr is used instead.
        :rtype: str
        """
        if litnamemap is True:
            litnamemap = self.get_litnamemap()
        return self.get_manager().sdd_to_dot(
            node=node, litnamemap=litnamemap, show_id=show_id, merge_leafs=merge_leafs
        )

    def get_litnamemap(self):
        """ Get a dictionary mapping literal IDs (inode index) to names. e.g; {1:'x', -1:'-x'}"""
        litnamemap = dict()
        var_count = self.get_manager().varcount
        for (name, index) in self.get_names():
            inode_index = self.atom2var.get(index, -1)
            if 0 <= inode_index < var_count:
                litnamemap[inode_index] = name
                litnamemap[-inode_index] = "-{}".format(name)
        return litnamemap

    def to_formula(self):
        """Extracts a LogicFormula from the SDD."""
        formula = LogicFormula(keep_order=True)

        for n, q, l in self.labeled():
            node = self.get_inode(q)
            constraints = self.get_constraint_inode()
            nodec = self.get_manager().conjoin(node, constraints)
            i = self._to_formula(formula, nodec, {})
            formula.add_name(n, i, l)
        return formula

    def _to_formula(self, formula, current_node, cache=None):
        if cache is not None and current_node.id in cache:
            return cache[current_node.id]
        if self.get_manager().is_true(current_node):
            retval = formula.TRUE
        elif self.get_manager().is_false(current_node):
            retval = formula.FALSE
        elif current_node.is_literal():  # it's a literal
            lit = current_node.literal
            at = self.var2atom[abs(lit)]
            node = self.get_node(at)
            if lit < 0:
                retval = -formula.add_atom(
                    -lit,
                    probability=node.probability,
                    name=node.name,
                    group=node.group,
                    cr_extra=False,
                    is_extra=node.is_extra,
                )
            else:
                retval = formula.add_atom(
                    lit,
                    probability=node.probability,
                    name=node.name,
                    group=node.group,
                    cr_extra=False,
                    is_extra=node.is_extra,
                )
        else:  # is decision
            # Formula: (p1^s1) v (p2^s2) v ...
            children = []
            for p, s in current_node.elements():
                p_n = self._to_formula(formula, p, cache)
                s_n = self._to_formula(formula, s, cache)
                c_n = formula.add_and((p_n, s_n))
                children.append(c_n)
            retval = formula.add_or(children)
        if cache is not None:
            cache[current_node.id] = retval
        return retval


class SDDManager(DDManager):
    """
    Manager for SDDs.
    It wraps around the SDD library and offers some additional methods.
    """

    def __init__(self, varcount=0, auto_gc=False, var_constraint=None):
        """Create a new SDD manager.

        :param varcount: number of initial variables
        :type varcount: int
        :param auto_gc: use automatic garbage collection and minimization
        :type auto_gc: bool
        :param var_constraint: A variable ordering constraint. Currently only x_constrained namedtuple are allowed.
        :type var_constraint: x_constrained
        """
        DDManager.__init__(self)
        if varcount is None or varcount <= 0:
            varcount = 1
        vtree = None
        if var_constraint is not None and varcount > 1:
            x_constraint = self._to_x_constrained_list(varcount, var_constraint)
            vtree = Vtree.new_with_X_constrained(
                var_count=varcount, is_X_var=x_constraint, vtree_type="balanced"
            )

        self.__manager = sdd.SddManager(
            var_count=varcount, auto_gc_and_minimize=auto_gc, vtree=vtree
        )
        self._assigned_varcount = 0

    def _to_x_constrained_list(self, varcount, var_constraint):
        """
        Convert the X-constrained var_constraint into a list of size varcount+1 specifying variables X. For variables i
        where 1 ≤ i ≤ varcount, if is_X_var[i] is 1 then i is in X, and if it is 0 then i is not in X.
        :param varcount: The amount of variables the vtree must have. This must be at least as high as the highest
            number in var_constraint.
        :param var_constraint: The X-constrained variable constraint
        :type var_constraint: x_constrained
        :return: The is_x_var input for the X-constrained vTree
        """
        is_x_var = [0] * (varcount + 1)
        if var_constraint is not None and len(var_constraint) > 0:
            for x in var_constraint.X:
                is_x_var[x] = 1
        return is_x_var

    def get_manager(self):
        """Get the underlying sdd manager."""
        return self.__manager

    @property
    def varcount(self):
        return self.get_manager().var_count()

    @property
    def assigned_varcount(self):
        return self._assigned_varcount

    def add_variable(self, label=0):
        if label == 0 or label > self.assigned_varcount:
            self._assigned_varcount += 1
            if self.assigned_varcount > self.varcount:
                self.get_manager().add_var_after_last()

            return self.assigned_varcount
        else:
            return label

    def literal(self, label):
        self.add_variable(abs(label))
        return self.get_manager().literal(label)

    def is_true(self, node):
        assert node is not None
        return node.is_true()

    def true(self):
        return self.get_manager().true()

    def is_false(self, node):
        assert node is not None
        return node.is_false()

    def false(self):
        return self.get_manager().false()

    def conjoin2(self, a, b):
        assert a is not None
        assert b is not None
        return self.get_manager().conjoin(a, b)

    def set_auto_gc_and_minimize(self, set_to=True):
        if set_to:
            self.__manager.auto_gc_and_minimize_on()
        else:
            self.__manager.auto_gc_and_minimize_off()

    def is_auto_gc_and_minimize_on(self):
        return self.__manager.is_auto_gc_and_minimize_on()

    def disjoin2(self, a, b):
        assert a is not None
        assert b is not None
        return self.get_manager().disjoin(a, b)

    def negate(self, node):
        assert node is not None
        new_sdd = self.get_manager().negate(node)
        self.ref(new_sdd)
        return new_sdd

    def same(self, node1, node2):
        # Assumes SDD library always reuses equivalent nodes.
        if node1 is None or node2 is None:
            return node1 == node2
        else:
            return node1.id == node2.id

    def ref(self, *nodes):
        for node in nodes:
            assert node is not None
            node.ref()

    def deref(self, *nodes):
        for node in nodes:
            assert node is not None
            node.deref()

    def write_to_dot(self, node, filename, litnamemap=None):
        if litnamemap is None:
            self.get_manager().save_as_dot(filename.encode(), node)
        else:
            with open(filename, "w") as file:
                file.write(self.sdd_to_dot(node=node, litnamemap=litnamemap))

    def to_internal_dot(self, node=None):
        """
        SDD for the given node, formatted for use with Graphviz dot.

        :param node: The node to get the dot from.
        :type node: SddNode
        :return: The dot format of the given node. When node is None, the shared_sdd will be used (contains all active
        sdd structures).
        :rtype: str
        """
        return self.get_manager().dot(node=node)

    def sdd_to_dot(self, node, litnamemap=None, show_id=False, merge_leafs=False):
        """
        SDD for the given node, formatted for use with Graphviz dot. This method provides more control over the used
        symbols than to_internal_dot (see litnamemap). Primes are given by a dotted line, subs by a full line.

        :param node: The node to get the dot from.
        :param litnamemap: A dictionary providing the symbols to use. The following options are available:
            1. literals, e.g. {1:'A', -1:'-A', ...},
            2. True/False, e.g. {true':'1', 'false':'0'}
            3. And/Or e.g. {'mult':'x', 'add':'+'}
        :type litnamemap: dict[(int | str), str] | None
        :param show_id: Whether to display the ids of each sdd node.
        :param merge_leafs: Whether to merge the same leaf nodes. True results in less nodes but makes it harder to
        render without having crossing lines.
        :return: The dot format of the given node. When node is None, the mgr is used (this behavior can be overriden).
        :rtype: str
        """
        used_node = node if node is not None else self.get_manager()
        return sdd_to_dot(
            node=used_node,
            litnamemap=litnamemap,
            show_id=show_id,
            merge_leafs=merge_leafs,
        )

    def wmc(
        self,
        node,
        weights,
        semiring,
        literal=None,
        pr_semiring=True,
        perform_smoothing=True,
        smooth_to_root=False,
        wmc_func=None,
    ):
        """Perform Weighted Model Count on the given node or the given literal.

        Common usage: wmc(node, weights, semiring) and wmc(node, weights, semiring, smooth_to_root=True)

        :param node: node to evaluate Type: SddNode
        :param weights: weights for the variables in the node. Type: {literal_id : (pos_weight, neg_weight)}
        :param semiring: use the operations defined by this semiring. Type: Semiring
        :param literal: When a literal is given, the result of WMC(literal) is returned instead.
        :param pr_semiring: Whether the given semiring is a (logspace) probability semiring.
        :param perform_smoothing: Whether to perform smoothing. When pr_semiring is True, smoothing is performed
            regardless.
        :param smooth_to_root: Whether to perform smoothing compared to the root. When pr_semiring is True, smoothing
            compared to the root is not performed regardless of this flag.
        :param wmc_func: The WMC function to use. If None, a built_in one will be used that depends on the given
            semiring. Type: function[SddNode, List[Tuple[prime_weight, sub_weight, Set[prime_used_lit],
            Set[sub_used_lit]]], Set[expected_prime_lit], Set[expected_sub_lit]] -> weight
        :type weights: dict[int, tuple[Any, Any]]
        :type semiring: Semiring
        :type pr_semiring: bool
        :type perform_smoothing: bool
        :type smooth_to_root: bool
        :type wmc_func: function
        :return: weighted model count of node if literal=None, else the weights are propagated up to node but the
            weighted model count of literal is returned.
        """
        varcount = self.get_manager().var_count()

        if pr_semiring and wmc_func is None:  # library built_in (WmcManager)
            logspace = 0
            if semiring.one() == 0.0:
                logspace = 1

            # setup
            wmc_manager = sdd.WmcManager(node, log_mode=logspace)
            for (
                n
            ) in weights:  # TODO wmc_manager.set_literal_weights_from_array is faster
                pos, neg = weights[n]
                if n <= varcount:
                    wmc_manager.set_literal_weight(n, pos)
                    wmc_manager.set_literal_weight(-n, neg)
            # Cover edge case e.g. node=SddNode(True)
            if varcount == 1 and weights.get(1) is None:
                wmc_manager.set_literal_weight(1, semiring.one())
                wmc_manager.set_literal_weight(-1, semiring.zero())

            # Calculate result
            result = wmc_manager.propagate()
            if literal is not None:
                result = wmc_manager.literal_pr(literal)
            if weights.get(0) is not None:  # Times the weight of True
                result = result * weights[0][0]
        else:  # manual iteration (SddIterator)
            if wmc_func is None:
                wmc_func = self._get_wmc_func(
                    weights=weights,
                    semiring=semiring,
                    perform_smoothing=perform_smoothing,
                )

            # Cover edge case e.g. node=SddNode(True)
            modified_weights = False
            if varcount == 1 and weights.get(1) is None:
                modified_weights = True
                weights[1] = (
                    semiring.one(),
                    semiring.zero(),
                )  # because 1 + 0 = 1 and 1 * x = x

            # Calculate result
            query_node = (
                node if literal is None else self.get_manager().literal(literal)
            )
            sdd_iterator = SddIterator(
                self.get_manager(), smooth_to_root=smooth_to_root
            )
            result = sdd_iterator.depth_first(query_node, wmc_func)

            if weights.get(0) is not None:  # Times the weight of True
                result = semiring.times(result, weights[0][0])

            # Restore edge case modification
            if modified_weights:
                weights.pop(1)

        self.get_manager().set_prevent_transformation(prevent=False)
        return result

    @staticmethod
    def _get_wmc_func(weights, semiring, perform_smoothing=True):
        """
        Get the function used to perform weighted model counting with the SddIterator. Smoothing supported.

        :param weights: The weights used during computations.
        :type weights: dict[int, tuple[Any, Any]]
        :param semiring: The semiring used for the operations.
        :param perform_smoothing: Whether smoothing must be performed. If false but semiring.is_nsp() then
            smoothing is still performed.
        :return: A WMC function that uses the semiring operations and weights, Performs smoothing if needed.
        """

        smooth_flag = perform_smoothing or semiring.is_nsp()

        def func_weightedmodelcounting(
            node, rvalues, expected_prime_vars, expected_sub_vars
        ):
            """ Method to pass on to SddIterator's ``depth_first`` to perform weighted model counting."""
            if rvalues is None:
                # Leaf
                if node.is_true():
                    result_weight = semiring.one()

                    # If smoothing, go over literals missed in scope
                    if smooth_flag:
                        missing_literals = (
                            expected_prime_vars
                            if expected_prime_vars is not None
                            else set()
                        )
                        missing_literals |= (
                            expected_sub_vars
                            if expected_sub_vars is not None
                            else set()
                        )

                        for missing_literal in missing_literals:
                            missing_pos_weight, missing_neg_weight = weights[
                                missing_literal
                            ]
                            missing_combined_weight = semiring.plus(
                                missing_pos_weight, missing_neg_weight
                            )
                            result_weight = semiring.times(
                                result_weight, missing_combined_weight
                            )

                    return result_weight

                elif node.is_false():
                    return semiring.zero()

                elif node.is_literal():
                    p_weight, n_weight = weights.get(abs(node.literal))
                    result_weight = p_weight if node.literal >= 0 else n_weight

                    # If smoothing, go over literals missed in scope
                    if smooth_flag:
                        lit_scope = {abs(node.literal)}

                        if expected_prime_vars is not None:
                            missing_literals = expected_prime_vars.difference(lit_scope)
                        else:
                            missing_literals = set()
                        if expected_sub_vars is not None:
                            missing_literals |= expected_sub_vars.difference(lit_scope)

                        for missing_literal in missing_literals:
                            missing_pos_weight, missing_neg_weight = weights[
                                missing_literal
                            ]
                            missing_combined_weight = semiring.plus(
                                missing_pos_weight, missing_neg_weight
                            )
                            result_weight = semiring.times(
                                result_weight, missing_combined_weight
                            )

                    return result_weight

                else:
                    raise Exception("Unknown leaf type for node {}".format(node))
            else:
                # Decision node
                if node is not None and not node.is_decision():
                    raise Exception("Expected a decision node for node {}".format(node))

                result_weight = None
                for prime_weight, sub_weight, prime_vars, sub_vars in rvalues:
                    branch_weight = semiring.times(prime_weight, sub_weight)

                    # If smoothing, go over literals missed in scope
                    if smooth_flag:
                        missing_literals = expected_prime_vars.difference(
                            prime_vars
                        ) | expected_sub_vars.difference(sub_vars)
                        for missing_literal in missing_literals:
                            missing_pos_weight, missing_neg_weight = weights[
                                missing_literal
                            ]
                            missing_combined_weight = semiring.plus(
                                missing_pos_weight, missing_neg_weight
                            )
                            branch_weight = semiring.times(
                                branch_weight, missing_combined_weight
                            )

                    # Add to current intermediate result
                    if result_weight is not None:
                        result_weight = semiring.plus(result_weight, branch_weight)
                    else:
                        result_weight = branch_weight

                return result_weight

        return func_weightedmodelcounting

    def wmc_literal(self, node, weights, semiring, literal):
        return self.wmc(node, weights, semiring, literal)

    def wmc_true(self, weights, semiring):
        return self.wmc(self.true(), weights, semiring)

    def count(self):
        return self.get_manager().count()

    def get_deepcopy_noref(
        self,
    ):  # TODO might be cleaner to maintain refcounts and deref everything afterwards in SDDExplicit
        """
        Get a deep copy of this without reference counts to inodes.
        Notes: No inode will have a reference count and auto_gc_and_minimize will be disabled.
        :return: A deep copy of this without any reference counts.
        """
        new_mgr = SDDManager(varcount=self.varcount, auto_gc=False)
        new_mgr._assigned_varcount = self.assigned_varcount

        # Code that is slower but works (see commented out code)
        mapping = self._copy_internal_SDD_to_noref(new_mgr=new_mgr, cache=None)

        # Construct node list based on current + mapping
        new_nodes = [None] * len(self.nodes)
        for i in range(0, len(self.nodes)):
            node = self.nodes[i]
            if node is not None:
                new_nodes[i] = mapping.get(self.nodes[i].id, None)

        new_mgr.nodes = new_nodes

        if self.constraint_dd is not None:
            new_mgr.constraint_dd = mapping.get(self.constraint_dd.id, None)

        # Old code which can be more efficient if we fix the "OSError Cannot allocate memory"
        # new_nodes = self.nodes.copy()
        # # fill in None spots with inode True
        # none_indices = []
        # true_node = self.true()
        # for i in range(0, len(new_nodes)):
        #     if new_nodes[i] is None:
        #         new_nodes[i] = true_node
        #         none_indices.append(i)
        # # push constraint node
        # if self.constraint_dd is not None:
        #     new_nodes.append(self.constraint_dd)
        #
        # # copy
        # new_mgr.__manager = self.get_manager().copy(new_nodes)
        #
        # # pop constraint node
        # if self.constraint_dd is not None:
        #     new_mgr.constraint_dd = new_nodes.pop()
        # # put back None
        # for i in none_indices:
        #     new_nodes[i] = None
        #
        # new_mgr.nodes = new_nodes

        # DEBUG - print before and after ref_counts
        # print("before %s" % [(node.ref_count() if node is not None else None) for node in self.nodes])
        # print("after %s" % [(node.ref_count() if node is not None else None) for node in new_mgr.nodes])

        return new_mgr

    def _copy_internal_SDD_to_noref(self, new_mgr, cache=None):
        """
        Copy the SDD structure of this manager to the given new_mgr. ALl refcounts will be 0 so the auto_gc on the
        new_mgr should be disabled.

        :param new_mgr: The manager to copy this SDD structure to. Note: the var_count must be high enough.
        :type new_mgr: SDDManager
        :param cache: The dictionary to use for the end result (mapping). If None, an empty one will be created.
        :type cache: dict[int, SddNode]
        :return: A mapping (cache) of {id, SddNode}. More specifically, {x:y} where x is the ID of an SddNode in this
            manager which corresponds to the newly created SddNode y of the new_mgr.
        :rtype: dict[int, SddNode]
        """
        if cache is None:
            cache = dict()
        root_nodes = self.nodes

        for root in root_nodes:
            if root is not None and root.ref_count() > 0:
                self._copy_internal_SDD_to_noref_aux(new_mgr, cache, root)
        return cache

    def _copy_internal_SDD_to_noref_aux(self, new_mgr, nodes_cache, node):
        """
        Copy node of the current SDDManager into new_mgr using nodes_cache to map (cache) SDDNode equivalences between
        the SDDNodes (id) of this manager to the SDDNodes in the new_mgr. This is an auxiliary method for
        self._copy_internal_SDD_to_noref(...).

        :param new_mgr: The new manager
        :type new_mgr: SDDManager
        :param nodes_cache: The cache to use to retrieve corresponding nodes between both managers. Has type
        {id: SddNode}. Must not be None.
        :type nodes_cache: dict[int, SddNode]
        :param node: The SddNode in this manager to copy over to new_mgr.
        :type node: SddNode
        :return: The SDDNode in new_mgr corresponding to node in this SDDManager.
        :rtype: SddNode
        """

        # cached_node: (SddNode x, SddNode y, int z)
        # with # x = node to process, y = x processed upto index z)
        cached_node = nodes_cache.get(node.id, None)
        if cached_node is not None:
            return cached_node

        stack = [(node, new_mgr.false(), -1)]
        while len(stack):
            process_node, inter_result, inter_index = stack.pop()

            if process_node.is_true():
                nodes_cache[process_node.id] = new_mgr.true()
                continue
            elif process_node.is_false():
                nodes_cache[process_node.id] = new_mgr.false()
                continue
            elif process_node.is_literal():
                new_node = new_mgr.literal(process_node.literal)
                nodes_cache[process_node.id] = new_node
                continue
            else:
                index = -1
                or_node = inter_result
                completed_for = True
                for p, s in process_node.elements():
                    # skip to inter_result
                    index = index + 1
                    if index <= inter_index:
                        continue

                    new_p = nodes_cache.get(p.id, None)
                    if new_p is None:
                        stack.append((process_node, or_node, index - 1))
                        stack.append((p, new_mgr.false(), -1))
                        completed_for = False
                        break

                    new_s = nodes_cache.get(s.id, None)
                    if new_s is None:
                        stack.append((process_node, or_node, index - 1))
                        stack.append((s, new_mgr.false(), -1))
                        completed_for = False
                        break

                    conjoined = new_mgr.conjoin(new_p, new_s)
                    or_node_n = new_mgr.disjoin(or_node, conjoined)
                    or_node_n.deref()  # ensure refcount = 0
                    conjoined.deref()  # ensure refcount = 0
                    # conjoined.deref()
                    # or_node.deref()
                    or_node = or_node_n
                if completed_for:
                    nodes_cache[process_node.id] = or_node

        return nodes_cache[node.id]

    def _copy_internal_SDD_to_aux_noref_rec(self, new_mgr, nodes_cache, node):
        """
        Copy node of the current SDDManager into new_mgr using nodes_cache to map (cache) SDDNode equivalences between
        the SDDNodes (id) of this manager to the SDDNodes in the new_mgr. This is an auxiliary method for
        self._copy_internal_SDD_to_noref(...) which uses recursion.

        :param new_mgr: The new manager
        :type new_mgr: SDDManager
        :param nodes_cache: The cache to use to retrieve corresponding nodes between both managers. Has type
        {id: SddNode}. Must not be None.
        :type nodes_cache: dict[int, SddNode]
        :param node: The SddNode in this manager to copy over to new_mgr.
        :type node: SddNode
        :return: The SDDNode in new_mgr corresponding to node in this SDDManager.
        :rtype: SddNode
        """
        cached = nodes_cache.get(node.id, None)
        if cached is not None:
            return cached

        if node.is_true():
            return new_mgr.true()
        elif node.is_false():
            return new_mgr.false()
        elif node.is_literal():
            new_node = new_mgr.literal(node.literal)
            nodes_cache[node.id] = new_node
            return new_node
        else:
            or_node = new_mgr.false()
            for p, s in node.elements():
                new_p = self._copy_internal_SDD_to_aux_noref_rec(
                    new_mgr, nodes_cache, p
                )
                new_s = self._copy_internal_SDD_to_aux_noref_rec(
                    new_mgr, nodes_cache, s
                )

                conjoined = new_mgr.conjoin(new_p, new_s)
                or_node_n = new_mgr.disjoin(or_node, conjoined)
                or_node_n.deref()  # ensure refcount = 0
                conjoined.deref()  # ensure refcount = 0
                # conjoined.deref()
                # or_node.deref()
                or_node = or_node_n
            nodes_cache[node.id] = or_node
            return or_node

    def __del__(self):
        # if sdd is not None and sdd.sdd_manager_free is not None:
        #     sdd.sdd_manager_free(self.__manager)
        self.__manager = None

    def __getstate__(self):
        tempfile = mktempfile()
        vtree = self.get_manager().vtree()  # not a copy
        vtree.save(tempfile.encode())
        with open(tempfile) as f:
            vtree_data = f.read()

        nodes = []
        for n in self.nodes:
            if n is not None:
                self.get_manager().save(tempfile.encode(), n)

                with open(tempfile) as f:
                    nodes.append(f.read())
            else:
                nodes.append(None)

        self.get_manager().save(tempfile.encode(), self.constraint_dd)
        with open(tempfile) as f:
            constraint_dd = f.read()

        os.remove(tempfile)
        return {
            "varcount": self.assigned_varcount,
            "nodes": nodes,
            "vtree": vtree_data,
            "constraint_dd": constraint_dd,
        }

    def __setstate__(self, state):
        self.nodes = []
        self._assigned_varcount = state["varcount"]
        tempfile = mktempfile()
        with open(tempfile, "w") as f:
            f.write(state["vtree"])
        vtree = sdd.Vtree.from_file(tempfile)
        self.__manager = sdd.SddManager.from_vtree(vtree)

        for n in state["nodes"]:
            if n is None:
                self.nodes.append(None)
            else:
                with open(tempfile, "w") as f:
                    f.write(n)
                self.nodes.append(self.__manager.read_sdd_file(tempfile.encode()))

        with open(tempfile, "w") as f:
            f.write(state["constraint_dd"])
        self.constraint_dd = self.__manager.read_sdd_file(tempfile.encode())
        os.remove(tempfile)
        return


x_constrained = namedtuple(
    "x_constrained", "X"
)  # X = list of literalIDs that have to appear before the rest


class SDDEvaluator(DDEvaluator):
    def __init__(self, formula, semiring, weights=None, **kwargs):
        DDEvaluator.__init__(self, formula, semiring, weights, **kwargs)

    def evaluate_custom(self, node):
        # Trivial case: node is deterministically True or False
        if node == self.formula.TRUE:
            if not self.semiring.is_nsp():
                result = self.semiring.one()
            else:
                # WMC(Theory & True & Evidence) / same. The constraints are already included in the evidence node.
                result = self.semiring.normalize(
                    self._evidence_weight, self._evidence_weight
                )
        elif node is self.formula.FALSE:
            result = self.semiring.zero()
        else:
            query_def_inode = self.formula.get_inode(node)
            evidence_inode = self.evidence_inode
            query_sdd = self._get_manager().conjoin(query_def_inode, evidence_inode)

            smooth_to_root = self.semiring.is_nsp()
            # perform_smoothing=True because of indicator variables # TODO There are no indicator variables in SDDs.
            result = self._get_manager().wmc(
                query_sdd,
                weights=self.weights,
                semiring=self.semiring,
                pr_semiring=False,
                perform_smoothing=True,
                smooth_to_root=smooth_to_root,
            )

            self._get_manager().deref(query_sdd)

            # TODO only normalize when there are evidence or constraints.
            #            result = self.semiring.normalize(result, self.normalization)
            result = self.semiring.normalize(result, self._evidence_weight)
        return self.semiring.result(result, self.formula)

    def _evaluate_evidence(self, recompute=False):
        if self._evidence_weight is None or recompute:
            constraint_inode = self.formula.get_constraint_inode()
            evidence_nodes = [self.formula.get_inode(ev) for ev in self.evidence()]
            self.evidence_inode = self._get_manager().conjoin(
                constraint_inode, *evidence_nodes
            )

            pr_semiring = isinstance(
                self.semiring, (SemiringProbability, SemiringLogProbability)
            )
            result = self._get_manager().wmc(
                self.evidence_inode,
                self.weights,
                self.semiring,
                pr_semiring=pr_semiring,
                perform_smoothing=True,
                smooth_to_root=False,
            )
            if self.semiring.is_zero(result):
                raise InconsistentEvidenceError(context=" during compilation")
            if self.normalization is None:
                self._evidence_weight = result
            else:
                self._evidence_weight = self.semiring.normalize(
                    result, self.normalization
                )

        return self._evidence_weight


@transform(LogicDAG, SDD)
def build_sdd(source, destination, **kwdargs):
    """Build an SDD from another formula.

    :param source: source formula
    :type source: LogicDAG
    :param destination: destination formula
    :type destination: SDD
    :param kwdargs: extra arguments
    :return: destination
    """
    init_varcount = kwdargs.get("init_varcount", -1)
    destination.init_varcount = (
        init_varcount if init_varcount != -1 else source.atomcount
    )
    return build_dd(source, destination, **kwdargs)



import os
from collections import defaultdict, namedtuple

from .errors import InvalidValue
from .logic import *
from .program import LogicProgram, PrologFile
from .util import OrderedSet


class ClauseDB(LogicProgram):
    """Compiled logic program.

    A logic program is compiled into a table of instructions.
    The types of instructions are:

    define( functor, arity, defs )
        Pointer to all definitions of functor/arity.
        Definitions can be: ``fact``, ``clause`` or ``adc``.

    clause( functor, arguments, bodynode, varcount )
        Single clause. Functor is the head functor, Arguments are the head arguments. Body node is a pointer to the node representing the body. Var count is the number of variables in head and body.

    fact( functor, arguments, probability )
        Single fact.

    adc( functor, arguments, bodynode, varcount, parent )
        Single annotated disjunction choice. Fields have same meaning as with ``clause``, parent_node points to the parent ``ad`` node.

    ad( childnodes )
        Annotated disjunction group. Child nodes point to the ``adc`` nodes of the clause.

    call( functor, arguments, defnode )
        Body literal with call to clause or builtin. Arguments contains the call arguments, definition node is the pointer to the definition node of the given functor/arity.

    conj( childnodes )
        Logical and. Currently, only 2 children are supported.

    disj( childnodes )
        Logical or. Currently, only 2 children are supported.

    neg( childnode )
        Logical not.

    """

    _define = namedtuple("define", ("functor", "arity", "children", "location"))
    _clause = namedtuple(
        "clause",
        (
            "functor",
            "args",
            "probability",
            "child",
            "varcount",
            "locvars",
            "group",
            "location",
        ),
    )
    _fact = namedtuple("fact", ("functor", "args", "probability", "location"))
    _call = namedtuple(
        "call", ("functor", "args", "defnode", "location", "op_priority", "op_spec")
    )
    _disj = namedtuple("disj", ("children", "location"))
    _conj = namedtuple("conj", ("children", "location"))
    _neg = namedtuple("neg", ("child", "location"))
    _choice = namedtuple(
        "choice",
        ("functor", "args", "probability", "locvars", "group", "choice", "location"),
    )
    _extern = namedtuple("extern", ("functor", "arity", "function"))

    FUNCTOR_CHOICE = "choice"
    FUNCTOR_BODY = "body"

    def __init__(self, builtins=None, parent=None):
        LogicProgram.__init__(self)
        self.__nodes = []  # list of nodes
        self.__heads = {}  # head.sig => node index

        self.__builtins = builtins

        self.data = {}
        self.engine = None

        self.__parent = parent
        self.__node_redirect = {}
        self.__extern = defaultdict(list)

        if parent is None:
            self.__offset = 0
        else:
            if hasattr(parent, "line_info"):
                self.line_info = parent.line_info
            if hasattr(parent, "source_files"):
                self.source_files = parent.source_files[:]
            self.__offset = len(parent)

        self.dont_cache = set()

        self.queries = []
        self._load_builtin_module()

    def _load_builtin_module(self):
        self.use_module(Term('library', Term('builtin')), None)
        # pass

    def __len__(self):
        return len(self.__nodes) + self.__offset

    def extend(self):
        return ClauseDB(parent=self, builtins=self.__builtins)

    def set_data(self, key, value):
        self.data[key] = value

    def update_data(self, key, value):
        if self.has_data(key):
            if type(value) == list:
                self.data[key] += value
            elif type(value) == dict:
                self.data[key].update(value)
            else:
                raise TypeError("Can't update data of type '%s'" % type(value))
        else:
            self.data[key] = value

    def has_data(self, key):
        return key in self.data

    def get_data(self, key, default=None):
        return self.data.get(key, default)

    def get_builtin(self, signature):
        if self.__builtins is None:
            if self.__parent is not None:
                return self.__parent.get_builtin(signature)
            else:
                return None
        else:
            return self.__builtins.get(signature)

    def get_reserved_names(self):
        return {self.FUNCTOR_CHOICE, self.FUNCTOR_BODY}

    def is_reserved_name(self, name):
        return name is self.get_reserved_names()

    def _create_index(self, arity):
        # return []
        return ClauseIndex(self, arity)

    def _add_and_node(self, op1, op2, location=None):
        """Add an *and* node."""
        return self._append_node(self._conj((op1, op2), location))

    def _add_not_node(self, op1, location=None):
        """Add a *not* node."""
        return self._append_node(self._neg(op1, location))

    def _add_or_node(self, op1, op2, location=None):
        """Add an *or* node."""
        return self._append_node(self._disj((op1, op2), location))

    def _scope_term(self, term, scope):
        if term.signature in self.__builtins:
            scope = None
        if term.functor == "_directive":
            scope = None
        if scope is not None:
            term.functor = "_%s_%s" % (scope, term.functor)
            return term
        else:
            return term

    def _add_define_node(self, head, childnode):
        define_index = self._add_head(head)
        define_node = self.get_node(define_index)
        if not define_node:
            clauses = self._create_index(head.arity)
            self._set_node(
                define_index,
                self._define(head.functor, head.arity, clauses, head.location),
            )
        else:
            clauses = define_node.children
        clauses.append(childnode)
        return childnode

    def _add_choice_node(
        self,
        choice,
        functor,
        args,
        probability,
        locvars,
        group,
        location=None,
        scope=None,
    ):
        choice_node = self._append_node(
            self._choice(functor, args, probability, locvars, group, choice, location)
        )
        return choice_node

    def _add_clause_node(self, head, body, varcount, locvars, group=None):
        clause_node = self._append_node(
            self._clause(
                head.functor,
                head.args,
                head.probability,
                body,
                varcount,
                locvars,
                group,
                head.location,
            )
        )
        return self._add_define_node(head, clause_node)

    def _add_call_node(self, term, scope=None):
        """Add a *call* node."""
        # if term.signature in ('query/1', 'evidence/1', 'evidence/2'):
        #    raise AccessError("Can\'t call %s directly." % term.signature)

        term = self._scope_term(term, scope)
        defnode = self._add_head(term, create=False)
        return self._append_node(
            self._call(
                term.functor,
                term.args,
                defnode,
                term.location,
                term.op_priority,
                term.op_spec,
            )
        )

    def get_node(self, index):
        """Get the instruction node at the given index.

        :param index: index of the node to retrieve
        :type index: :class:`int`
        :returns: requested node
        :rtype: :class:`tuple`
        :raises IndexError: the given index does not point to a node
        """
        index = self.__node_redirect.get(index, index)

        if index < self.__offset:
            return self.__parent.get_node(index)
        else:
            return self.__nodes[index - self.__offset]

    def _set_node(self, index, node):
        if index < self.__offset:
            raise IndexError("Can't update node in parent.")
        else:
            self.__nodes[index - self.__offset] = node

    def _append_node(self, node=()):
        index = len(self)
        self.__nodes.append(node)
        return index

    def _get_head(self, head=None):
        node = self.__heads.get(head.signature)
        if node is None and self.__parent:
            node = self.__parent._get_head(head)
        return node

    def _set_head(self, head, index):
        self.__heads[head.signature] = index

    def _add_head(self, head, create=True):
        if self.is_reserved_name(head.functor):
            raise AccessError("'%s' is a reserved name" % head.functor)
        node = self.get_builtin(head.signature)
        if node is not None:
            if create:
                raise AccessError("Can not overwrite built-in '%s'." % head.signature)
            else:
                return node

        node = self._get_head(head)
        if node is None:
            if create:
                node = self._append_node(
                    self._define(
                        head.functor,
                        head.arity,
                        self._create_index(head.arity),
                        head.location,
                    )
                )
            else:
                node = self._append_node()
            self._set_head(head, node)
        elif create and node < self.__offset:
            existing = self.get_node(node)
            # node exists in parent
            clauses = self._create_index(head.arity)
            if existing:
                for c in existing.children:
                    clauses.append(c)
            old_node = node
            node = self._append_node(
                self._define(head.functor, head.arity, clauses, head.location)
            )
            self.__node_redirect[old_node] = node
            self._set_head(head, node)

        return node

    def find(self, head):
        """Find the ``define`` node corresponding to the given head.

        :param head: clause head to match
        :type head: :class:`.basic.Term`
        :returns: location of the clause node in the database, \
                     returns ``None`` if no such node exists
        :rtype: :class:`int` or ``None``
        """
        return self._get_head(head)

    def __repr__(self):
        s = ""
        for i, n in enumerate(self.__nodes):
            i += self.__offset
            s += "%s: %s\n" % (i, n)
        s += str(self.__heads)
        s += "\n"
        s += "Redirects: " + str(self.__node_redirect)
        return s

    def add_clause(self, clause, scope=None):
        """Add a clause to the database.

       :param clause: Clause to add
       :type clause: Clause
       :returns: location of the definition node in the database
       :rtype: int
        """
        return self._compile(clause, scope=scope)

    def add_fact(self, term, scope=None):
        """Add a fact to the database.
       :param term: fact to add
       :type term: Term
       :return: position of the definition node in the database
       :rtype: int
        """
        # Count the number of variables in the fact
        variables = _AutoDict()
        term.apply(variables)
        # If the fact has variables, threat is as a clause.
        if len(variables) == 0:
            term = self._scope_term(term, scope)
            fact_node = self._append_node(
                self._fact(term.functor, term.args, term.probability, term.location)
            )
            return self._add_define_node(term, fact_node)
        else:
            return self.add_clause(
                Clause(term, Term("true")),
                scope=scope,
            )

    def add_extern(self, predicate, arity, func, scope=None):
        head = Term(predicate, *[None] * arity)
        head = self._scope_term(head, scope)
        node_id = self._get_head(head)
        ext = self._extern(head.functor, head.arity, func)
        if node_id is None:
            node_id = self._append_node(ext)
            self._set_head(head, node_id)
        else:
            node = self.get_node(node_id)
            if node == ():
                self._set_node(node_id, ext)
            else:
                node_id = self._append_node(ext)
                self._add_define_node(head, node_id)
        self.__extern[scope].append(Term("'/'", Term(predicate), Constant(arity)))

    def get_local_scope(self, signature):
        if signature in ("findall/3", "all/3", "all_or_none/3"):
            return 0, 1
        else:
            return []

    def _compile(self, struct, variables=None, scope=None):
        """Compile the given structure and add it to the database.

        :param struct: structure to compile
        :type struct: Term
        :param variables: mapping between variable names and variable index
        :type variables: _AutoDict
        :return: position of the compiled structure in the database
        :rtype: int
        """
        if variables is None:
            variables = _AutoDict()

        if isinstance(struct, And):
            op1 = self._compile(
                struct.op1, variables, scope=scope
            )
            op2 = self._compile(
                struct.op2, variables, scope=scope
            )
            return self._add_and_node(op1, op2)
        elif isinstance(struct, Or):
            op1 = self._compile(
                struct.op1, variables, scope=scope
            )
            op2 = self._compile(
                struct.op2, variables, scope=scope
            )
            return self._add_or_node(op1, op2)
        elif isinstance(struct, Not):
            variables.enter_local()
            child = self._compile(
                struct.child, variables, scope=scope
            )
            variables.exit_local()
            return self._add_not_node(child, location=struct.location)
        elif isinstance(struct, Term) and struct.signature == "not/1":
            child = self._compile(
                struct.args[0],
                variables,
                scope=scope,
            )
            return self._add_not_node(child, location=struct.location)
        elif isinstance(struct, AnnotatedDisjunction):
            # Determine number of variables in the head
            new_heads = [head.apply(variables) for head in struct.heads]

            # Group id
            group = len(self.__nodes)

            # Create the body clause
            body_node = self._compile(
                struct.body, variables, scope=scope
            )
            body_count = len(variables)
            # Body arguments
            body_args = tuple(range(0, len(variables)))
            body_functor = self.FUNCTOR_BODY + "_" + str(len(self))
            if len(new_heads) > 1:
                heads_list = Term("multi")  # list2term(new_heads)
            else:
                heads_list = new_heads[0].with_probability(None)
            body_head = Term(body_functor, Constant(group), heads_list, *body_args)
            self._add_clause_node(
                body_head, body_node, len(variables), variables.local_variables
            )
            clause_body = self._add_head(body_head)
            for choice, head in enumerate(new_heads):
                head = self._scope_term(head, scope)
                # For each head: add choice node
                choice_functor = Term(
                    self.FUNCTOR_CHOICE,
                    Constant(group),
                    Constant(choice),
                    head.with_probability(),
                )
                choice_node = self._add_choice_node(
                    choice,
                    choice_functor,
                    body_args,
                    head.probability,
                    variables.local_variables,
                    group,
                    head.location,
                )
                choice_call = self._append_node(
                    self._call(
                        choice_functor,
                        body_args,
                        choice_node,
                        head.location,
                        None,
                        None,
                    )
                )
                body_call = self._append_node(
                    self._call(
                        body_functor,
                        body_head.args,
                        clause_body,
                        head.location,
                        None,
                        None,
                    )
                )
                choice_body = self._add_and_node(body_call, choice_call)
                self._add_clause_node(head, choice_body, body_count, {}, group=group)
            return None
        elif isinstance(struct, Clause):
            if struct.head.probability is not None:
                return self._compile(
                    AnnotatedDisjunction([struct.head], struct.body),
                    scope=scope,
                )
            else:
                new_head = self._scope_term(
                    struct.head.apply(variables),
                    scope,
                )
                body_node = self._compile(
                    struct.body,
                    variables,
                    scope=scope,
                )
                return self._add_clause_node(
                    new_head, body_node, len(variables), variables.local_variables
                )
        elif isinstance(struct, Var):
            return self._add_call_node(
                Term("call", struct.apply(variables), location=struct.location),
                scope=scope,
            )
        elif isinstance(struct, Term):
            local_scope = self.get_local_scope(struct.signature)
            if local_scope:
                # Special case for findall: any variables added by the first
                #  two arguments of findall are 'local' variables.
                args = []
                for i, a in enumerate(struct.args):
                    if not isinstance(a, Term):
                        # For nested findalls: 'a' can be a raw variable pointer
                        # Temporarily wrap it in a Term, so we can call 'apply' on it.
                        a = Term("_", a)
                    if i in local_scope:
                        variables.enter_local()
                        new_arg = a.apply(variables)
                        variables.exit_local()
                    else:
                        new_arg = a.apply(variables)
                    if a.functor == "_":
                        # If the argument was temporarily wrapped: unwrap it.
                        new_arg = new_arg.args[0]
                    args.append(new_arg)
                return self._add_call_node(
                    struct(*args), scope=scope
                )
            elif struct.functor in ("consult", "use_module"):
                new_struct = Term(
                    "_" + struct.functor,
                    Term(scope),
                    *struct.args,
                    location=struct.location
                )
                return self._add_call_node(
                    new_struct.apply(variables),
                    scope=scope,
                )
            else:
                return self._add_call_node(
                    struct.apply(variables),
                    scope=scope,
                )
        else:
            raise ValueError("Unknown structure type: '%s'" % struct)

    def _create_ints(self, term):
        """
        Convert all occurrences of Var into ints.

        For example, when term = Term('a', Var('V_1'), Constant(3)), this returns Term('a', 1, Constant(3))
        This assumes each Var has functor V_{var_id} with var_id some int value.
        This is opposite to the method _create_vars(term).
        :param term: The term to replace all Var occurrences in.
        :type term: Term
        :return: The same term but with all occurrences of Var replaced with their corresponding int value.
        :rtype: Term
        """
        if type(term) == Var:
            assert term.functor[:2] == "V_"
            # Convert V_{digit} to an int of value {digit}
            return int(term.functor[2:])
        else:
            args = [self._create_ints(arg) for arg in term.args]
            term = term.with_args(*args)
            if term.probability is not None:
                term = term.with_probability(self._create_ints(term.probability))
            return term

    def _create_vars(self, term):
        if type(term) == int:
            return Var("V_" + str(term))
        else:
            args = [self._create_vars(arg) for arg in term.args]
            term = term.with_args(*args)
            if term.probability is not None:
                term = term.with_probability(self._create_vars(term.probability))
            return term

    def _extract(self, node_id):
        node = self.get_node(node_id)
        if not node:
            raise ValueError("Unexpected empty node.")

        nodetype = type(node).__name__
        if nodetype == "fact":
            return Term(node.functor, *node.args, p=node.probability)
        elif nodetype == "call":
            func = node.functor
            args = node.args
            if isinstance(func, Term):
                return self._create_vars(func(*(func.args + args)))
            else:
                return self._create_vars(
                    Term(func, *args, priority=node.op_priority, opspec=node.op_spec)
                )
        elif nodetype == "conj":
            a, b = node.children
            return And(self._extract(a), self._extract(b))
        elif nodetype == "disj":
            a, b = node.children
            return Or(self._extract(a), self._extract(b))
        elif nodetype == "neg":
            return Not("\\+", self._extract(node.child))
        else:
            raise ValueError("Unknown node type: '%s'" % nodetype)

    def to_clause(self, index):
        node = self.get_node(index)
        nodetype = type(node).__name__
        if nodetype == "fact":
            return Term(node.functor, *node.args, p=node.probability)
        elif nodetype == "clause":
            head = self._create_vars(Term(node.functor, *node.args, p=node.probability))
            return Clause(head, self._extract(node.child))

    def __iter__(self):
        clause_groups = defaultdict(list)
        for index, node in self.enum_nodes():
            if not node:
                continue
            nodetype = type(node).__name__
            if nodetype == "fact":
                yield Term(node.functor, *node.args, p=node.probability)
            elif nodetype == "clause":
                if node.group is None:
                    head = self._create_vars(
                        Term(node.functor, *node.args, p=node.probability)
                    )
                    yield Clause(head, self._extract(node.child))
                else:
                    clause_groups[node.group].append(index)
        for group in clause_groups.values():
            heads = []
            body = None
            for index in group:
                node = self.get_node(index)
                heads.append(
                    self._create_vars(
                        Term(node.functor, *node.args, p=node.probability)
                    )
                )
                if body is None:
                    body_node = self.get_node(node.child)
                    body_node = self.get_node(body_node.children[0])
                    body = self._create_vars(Term(body_node.functor, *body_node.args))
            yield AnnotatedDisjunction(heads, body)

    def iter_raw(self):
        """Iterate over clauses of model as represented in the database i.e. with choice facts and
         without annotated disjunctions.
        """

        clause_groups = defaultdict(list)
        for index, node in self.enum_nodes():
            if not node:
                continue
            nodetype = type(node).__name__
            if nodetype == "fact":
                yield Term(node.functor, *node.args, p=node.probability)
            elif nodetype == "clause":
                if node.group is None:
                    head = self._create_vars(
                        Term(node.functor, *node.args, p=node.probability)
                    )
                    yield Clause(head, self._extract(node.child))
                else:
                    head = self._create_vars(Term(node.functor, *node.args))
                    yield Clause(head, self._extract(node.child))
            elif nodetype == "choice":
                group = node.functor.args[0]
                c = node.functor(*(node.functor.args + node.args))
                clause_groups[group].append(c)
                yield c.with_probability(node.probability)

        for group in clause_groups.values():
            if len(group) > 1:
                yield Term("mutual_exclusive", list2term(group))

    def resolve_filename(self, filename):
        if (
            hasattr(filename, "functor")
            and filename.functor == "library"
            and filename.arity == 1
        ):
            from . import library_paths

            libname = unquote(str(filename.args[0]))
            for path in library_paths:
                filename = os.path.join(path, libname)
                if os.path.exists(filename):
                    return filename
                elif os.path.exists(filename + ".pl"):
                    return filename + ".pl"
                elif os.path.exists(filename + ".py"):
                    return filename + ".py"
        else:
            root = self.source_root
            if hasattr(filename, "location") and filename.location:
                source_root = self.source_files[filename.location[0]]
                if source_root:
                    root = os.path.dirname(source_root)

            filename = os.path.join(root, unquote(str(filename)))
            if os.path.exists(filename):
                return filename
            elif os.path.exists(filename + ".pl"):
                return filename + ".pl"
            elif os.path.exists(filename + ".py"):
                return filename + ".py"
        return filename

    def create_function(self, functor, arity):
        """Create a Python function that can be used to query a specific predicate on this database.

        :param functor: functor of the predicate
        :param arity: arity of the predicate (the function will take arity - 1 arguments
        :return: a Python callable
        """
        return PrologFunction(self, functor, arity)

    def enum_nodes(self):
        if self.__parent:
            for i, n in self.__parent.enum_nodes():
                yield i, n
        for i, n in enumerate(self.__nodes):
            i += self.__offset
            yield i, n

    def iter_nodes(self):
        if self.__parent:
            for n in self.__parent.iter_nodes():
                yield n
        for n in self.__nodes:
            yield n

    def consult(self, filename, location=None, my_scope=None):
        filename = self.resolve_filename(filename)
        if filename is None:
            raise ConsultError(
                message="Consult: file not found '%s'" % filename,
                location=self.lineno(location),
            )

        # Prevent loading the same file twice
        if filename not in self.source_files:
            identifier = len(self.source_files)
            self.source_files.append(filename)
            self.source_parent.append(location)
            program = PrologFile(
                filename,
                identifier=identifier,
                factory=self.extra_info.get("factory"),
                parser=self.extra_info.get("parser"),
            )
            self.line_info.append(program.line_info[0])
            # engine._process_directives(database)
            return self.add_all(program)
        else:
            return None, None

    def add_all(self, program):
        module_name = None
        module_preds = None
        for index, clause in enumerate(program):
            if (
                clause.functor == ":-"
                and hasattr(clause.args[0], "functor")
                and clause.args[0].functor == "_directive"
                and clause.args[1].signature == "module/2"
            ):
                if index > 0:
                    raise AccessError(
                        "'module' directive should appear at top of module"
                    )
                module_name = str(clause.args[1].args[0])
                module_preds = clause.args[1].args[1]
            else:
                self.add_statement(clause, module_name)
        if module_preds is not None:
            module_preds = term2list(module_preds)
        return module_name, module_preds

    def use_module(self, filename, predicates, location=None, my_scope=None):
        filename = self.resolve_filename(filename)
        if filename is None:
            raise ConsultError("Unknown library location", self.lineno(location))
        elif filename is not None and filename[-3:] == ".py":
            try:
                module_name, module_predicates = self.load_external_module(filename)
            except IOError as err:
                raise ConsultError(
                    "Error while reading external library: %s" % str(err),
                    self.lineno(location),
                )
        else:
            module_name, module_predicates = self.consult(
                Term(filename), location=location
            )

        if module_name is not None:
            if predicates is None:
                for mp in module_predicates:
                    self._create_alias(mp, module_name, my_scope=my_scope)
            elif predicates.functor == "except":
                preds = set(term2list(predicates.args[0]))
                for mp in module_predicates:
                    if mp not in preds:
                        self._create_alias(mp, module_name, my_scope=my_scope)
            else:
                preds = {}
                for pred in term2list(predicates):
                    if pred.functor == "'as'":
                        mp = pred.args[0]
                        rename = pred.args[1]
                    else:
                        mp = pred
                        rename = pred.args[0]
                    if mp in module_predicates:
                        self._create_alias(
                            mp, module_name, rename=rename, my_scope=my_scope
                        )
                    else:
                        raise GroundingError(
                            "Imported predicate %s not defined in module %s"
                            % (mp, module_name),
                            location=location,
                        )

    def _create_alias(
        self, pred, scope, rename=None, my_scope=None
    ):
        if rename is None:
            rename = pred.args[0]

        if scope is not None:
            root_sign = self._scope_term(
                Term(rename, *[None] * int(pred.args[1])),
                my_scope,
            )
            scoped_sign = self._scope_term(
                Term(pred.args[0], *[None] * int(pred.args[1])),
                scope,
            )

            rh = self._add_head(root_sign, create=False)
            sh = self._add_head(scoped_sign, create=False)

            if self.get_node(rh):
                # TODO warning that user code overrides library
                pass
            elif rh is not None:
                sh = self.__node_redirect.get(sh, sh)
                self.__node_redirect[rh] = sh

    def load_external_module(self, filename):
        from .extern import problog_export

        problog_export.database = self
        module_name = os.path.splitext(os.path.split(filename)[-1])[0]

        # Replaced deprecated imp.load_module()
        # source: https://stackoverflow.com/a/19011259
        import importlib.machinery
        import importlib.util
        loader = importlib.machinery.SourceFileLoader(module_name, filename)
        module_spec = importlib.util.spec_from_loader(loader.name, loader)
        new_module = importlib.util.module_from_spec(module_spec)
        loader.exec_module(new_module)

        return module_name, self.__extern[module_name]


class ConsultError(GroundingError):
    """Error during consult"""

    def __init__(self, message, location):
        GroundingError.__init__(self, message, location)


def _atom_to_filename(atom):
    """Translate an atom to a filename.

   :param atom: filename as atom
   :type atom: Term
   :return: filename as string
   :rtype: str
    """
    atomstr = str(atom)
    if atomstr[0] == atomstr[-1] == "'":
        atomstr = atomstr[1:-1]
    return atomstr


class PrologFunction(object):
    def __init__(self, database, functor, arity):
        self.database = database
        self.functor = functor
        self.arity = arity

    def __call__(self, *args):
        args = args[: self.arity - 1]
        query_term = Term(self.functor, *(args + (None,)))
        result = self.database.engine.query(self.database, query_term)
        if len(result) != 1:
            raise InvalidValue(
                "Function should return one result: %s returned %s"
                % (query_term, result)
            )
        return result[0][-1]


class AccessError(GroundingError):
    pass


class _AutoDict(dict):
    def __init__(self):
        dict.__init__(self)
        self.__record = set()
        self.__anon = 0
        self.__localmode = False
        self.local_variables = set()

    def enter_local(self):
        self.__localmode = True

    def exit_local(self):
        self.__localmode = False

    def __getitem__(self, key):
        if key == "_" and self.__localmode:
            key = "_#%s" % len(self.local_variables)

        if key == "_" or key is None:

            value = len(self)
            self.__anon += 1
            return value
        else:
            value = self.get(key)
            if value is None:
                value = len(self)
                self[key] = value
                if self.__localmode:
                    self.local_variables.add(value)
            elif not self.__localmode and value in self.local_variables:
                # Variable initially defined in local scope is reused outside local scope.
                # This means it's not local anymore.
                self.local_variables.remove(value)
            self.__record.add(value)
            return value

    def __len__(self):
        return dict.__len__(self) + self.__anon

    def usedVars(self):
        result = set(self.__record)
        self.__record.clear()
        return result

    def define(self, key):
        if key not in self:
            value = len(self)
            self[key] = value


def intersection(l1, l2):
    i = 0
    j = 0
    n1 = len(l1)
    n2 = len(l2)
    r = []
    a = r.append
    while i < n1 and j < n2:
        if l1[i] == l2[j]:
            a(l1[i])
            i += 1
            j += 1
        elif l1[i] < l2[j]:
            i += 1
        else:
            j += 1
    return r


class ClauseIndex(list):
    def __init__(self, parent, arity):
        list.__init__(self)
        self.__parent = parent
        self.__basetype = OrderedSet
        self.__index = [defaultdict(self.__basetype) for _ in range(0, arity)]
        self.__optimized = False
        self.__erased = set()

    def find(self, arguments):
        results = None
        for i, arg in enumerate(arguments):
            if not is_ground(arg):
                pass  # Variable => no restrictions
            else:
                curr = self.__index[i].get(arg)
                none = self.__index[i].get(None, self.__basetype())
                if curr is None:
                    curr = none
                else:
                    curr |= none

                if results is None:  # First argument with restriction
                    results = curr
                else:
                    results = results & curr  # for some reason &= doesn't work here
            if results is not None and not results:
                return []
        if results is None:
            if self.__erased:
                return OrderedSet(self) - self.__erased
            else:
                return self
        else:
            if self.__erased:
                return results - self.__erased
            else:
                return results

    def _add(self, key, item):
        for i, k in enumerate(key):
            self.__index[i][k].add(item)

    def append(self, item):
        list.append(self, item)
        key = []
        try:
            args = self.__parent.get_node(item).args
        except AttributeError:
            args = [None] * self.__parent.get_node(item).arity
        for arg in args:
            if is_ground(arg):
                key.append(arg)
            else:
                key.append(None)
        self._add(key, item)

    def erase(self, items):
        self.__erased |= set(items)



import unittest

from problog.logic import Term
from problog.pypl import py2pl, pl2py


class TestPypl(unittest.TestCase):
    def test_list_double_conversion(self):
        test_list = [1, 2, 3, "a", (4, 5, "c"), 7.0, [8, "d"], Term("h")]
        res_list = pl2py(py2pl(test_list))
        self.assertEqual(test_list, res_list)

    def test_tuple_double_conversion(self):
        test_tuple = (1, 2, 3, "a", [4, 5, "c"], 7.0, [8, "d"])
        res_tuple = pl2py(py2pl(test_tuple))
        self.assertEqual(test_tuple, res_tuple)

    def test_empty_list_double_conversion(self):
        test_list = []
        res_list = pl2py(py2pl(test_list))
        self.assertEqual(test_list, res_list)

    def test_empty_tuple_double_conversion(self):
        test_tuple = ()
        res_tuple = pl2py(py2pl(test_tuple))
        self.assertEqual(test_tuple, res_tuple)

    def test_unsupported_types(self):
        test_dict = {"a": 1}
        with self.assertRaises(ValueError):
            py2pl(test_dict)

    def test_list_at_end_of_comma_double_conversion(self):
        test_value = (1, [])
        res_value = pl2py(py2pl(test_value))
        self.assertEqual(test_value, res_value)



"""
problog.setup - Installation tools
----------------------------------

Provides an installer for ProbLog dependencies.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import shutil

import os
import subprocess
import sys


def get_system():
    system = sys.platform
    if system.lower().startswith("java"):
        import java.lang.System

        system = java.lang.System.getProperty("os.name").lower()
    if system.startswith("linux"):
        system = "linux"
    elif system.startswith("win"):
        system = "windows"
    elif system.startswith("mac"):
        system = "darwin"
    return system


def set_environment():
    """Updates local PATH and PYTHONPATH to include additional component directories."""
    # Set PATH environment
    path = os.environ.get("PATH", [])
    if path:
        path = [path]
    path = os.pathsep.join(path + list(get_binary_paths()))
    os.environ["PATH"] = path
    # Set PYTHONPATH environment
    sys.path += get_module_paths()


def get_binary_paths():
    """Get a list of additional binary search paths."""
    binary_root = os.path.join(os.path.dirname(__file__), "bin")
    system = get_system()  # Darwin, Linux, ?
    return list(map(os.path.abspath, [os.path.join(binary_root, system), binary_root]))


def get_module_paths():
    """Get a list of additional module search paths."""
    binary_root = os.path.join(os.path.dirname(__file__), "lib")
    system = get_system()  # Darwin, Linux, ?
    python = "python%s" % sys.version_info[0]
    return list(
        map(
            os.path.abspath,
            [
                os.path.join(binary_root, python, system),
                os.path.join(binary_root, python),
                binary_root,
            ],
        )
    )


def gather_info():
    """Collect info about the system and its installed software."""

    system_info = {}

    system_info["root_path"] = os.path.join(os.path.dirname(__file__), "..")

    system_info["os"] = get_system()
    # system_info['arch'] = os.uname()[-1]

    system_info["python_version"] = sys.version_info

    # Module pyparsing
    try:
        import pyparsing

        system_info["pyparsing"] = pyparsing.__version__
    except ImportError:
        pass

    # SDD module
    # noinspection PyBroadException
    try:
        import pysdd

        system_info["sdd_module"] = True
    except Exception:
        pass

    # DSharp
    system_info["dsharp"] = shutil.which("dsharp") is not None

    # c2d
    system_info["c2d"] = shutil.which("cnf2dDNNF") is not None
    return system_info


def detect_compiler():
    """Detects the available C compiler."""
    if shutil.which("gcc"):
        return "gcc"
    elif shutil.which("clang"):
        return "clang"
    elif shutil.which("cl"):  # Microsoft Visual C++ (MSVC)
        return "cl"
    else:
        raise RuntimeError("No suitable C compiler found")

def build_maxsatz():
    if get_system() == "windows":
        return  # We include the binary

    dest_dir, source_dir = get_binary_paths()
    source_dir = os.path.join(source_dir, "source", "maxsatz")
    source_file = "maxsatz2009.c"
    output_file = os.path.join(dest_dir, "maxsatz")

    compiler = detect_compiler()
    if compiler not in ["gcc", "clang"]:
        return

    with WorkingDir(source_dir):
        compile_cmd = [compiler, "-o", output_file, source_file]
        subprocess.run(compile_cmd, check=True)


def install(force=True):
    info = gather_info()
    build_maxsatz()
    return info


def system_info():
    info = gather_info()

    ok = True
    s = "System information:\n"
    s += "------------------:\n"
    s += "Operating system: %s\n" % info.get("os", "unknown")
    s += "System architecture: %s\n" % info.get("arch", "unknown")
    s += "Python version: %s.%s.%s\n" % (
        info["python_version"].major,
        info["python_version"].minor,
        info["python_version"].micro,
    )
    s += "\n"
    s += "ProbLog components:\n"
    s += "-------------------\n"

    # PrologFile, PrologString  => require pyparsing
    # SDD => requires sdd_library
    # NNF => requires dsharp or c2d

    # SemiringOther => requires NNF (or SDD with alternative evaluation)

    # pyparsing = info.get('pyparsing', 'NOT INSTALLED')
    # s += 'Module \'pyparsing\': %s\n' % pyparsing
    # if not pyparsing :
    #     s += '  ACTION: install the pyparsing module\n'
    # sdd = info.get('sdd_module', False)
    # if sdd :
    #     s += 'Module \'sdd\': INSTALLED\n'
    # else :
    #     s += 'Module \'sdd\': NOT INSTALLED\n'
    #     s += '  ACTION: run ProbLog installer\n'
    #
    return s


class WorkingDir(object):
    def __init__(self, workdir):
        self.workdir = workdir
        self.currentdir = os.path.abspath(os.curdir)

    def __enter__(self):
        self.currentdir = os.path.abspath(os.curdir)
        os.chdir(self.workdir)

    def __exit__(self, *args):
        os.chdir(self.currentdir)


if __name__ == "__main__":
    set_environment()
    info = install()
    print(info)



#! /usr/bin/env python
"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import os
import sys

sys.setrecursionlimit(10000)

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))

from problog.errors import ProbLogError


def print_result_prob(d, output, precision=8):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    result = {}
    success, d = d
    if success:
        result["SUCCESS"] = True
        result["probs"] = [
            [str(n), round(p, precision), n.loc[1], n.loc[2]] for n, p in d.items()
        ]
    else:
        result["SUCCESS"] = False
        result["err"] = process_error(d)
    print(200, "application/json", json.dumps(result), file=output)
    return 0


def print_result_mpe(d, output, precision=8):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    result = {}
    success, d = d
    if success:
        result["SUCCESS"] = True
        result["atoms"] = list(
            map(lambda n: (str(-n), False) if n.is_negated() else (str(n), True), d)
        )
    else:
        result["SUCCESS"] = False
        result["err"] = process_error(d)
    print(200, "application/json", json.dumps(result), file=output)
    return 0


def print_result_sample(d, output, **kwdargs):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    result = {}
    success, d = d
    if success:
        result["SUCCESS"] = True
        result["results"] = [[(str(k), str(v)) for k, v in dc.items()] for dc in d]
    else:
        result["SUCCESS"] = False
        result["err"] = process_error(d)
    print(200, "application/json", json.dumps(result), file=output)
    return 0


def process_error(err):
    if isinstance(err, ProbLogError):
        return vars(err)
    else:
        return {"message": "An unexpected error has occurred (%s)." % err}


def main(args):
    task = args[0]
    args = list(args[:-1]) + ["-o", args[-1]]
    from problog.tasks import run_task

    if task == "mpe":
        print_result = print_result_mpe
    elif task == "sample":
        print_result = print_result_sample
    else:
        print_result = print_result_prob

    run_task(args, print_result)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
problog.sdd_formula_explicit - Sentential Decision Diagrams
--------------------------------------------------

Interface to Sentential Decision Diagrams (SDD) using the explicit encoding representing all models
(similar to d-DNNF encoding except that it is not converted into cnf first).

..
    Part of the ProbLog distribution.

    Copyright 2018 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from collections import namedtuple

from .core import transform
from .errors import InconsistentEvidenceError
from .evaluator import SemiringProbability, SemiringLogProbability
from .formula import LogicDAG, LogicFormula
from .forward import _ForwardSDD, ForwardInference
from .sdd_formula import SDDEvaluator, SDD, x_constrained
from .sdd_formula import SDDManager
from .util import Timer

# noinspection PyBroadException
try:
    from pysdd import sdd
    from pysdd.iterator import SddIterator
except Exception as err:
    sdd = None


class SDDExplicit(SDD):
    """
        This formula is using the cnf-encoding (c :- a,b = {c,a,b} v {-c,(-a v -b)}). This implies there is an
        indicator variable for each derived literal and the circuit consists of a single root node on which we do WMC.
        Evidence and querying is done by modifying the weights.
    """

    transform_preference = 11

    def __init__(self, sdd_auto_gc=False, **kwdargs):
        SDD.__init__(self, sdd_auto_gc=sdd_auto_gc, **kwdargs)
        self._root = None

    def _create_manager(self):
        return SDDExplicitManager(
            auto_gc=self.auto_gc,
            var_constraint=self.var_constraint,
            varcount=self.init_varcount,
        )

    def _create_evaluator(self, semiring, weights, **kwargs):
        return SDDExplicitEvaluator(self, semiring, weights, **kwargs)

    def get_root_inode(self):
        """ Get the current root inode. This includes the constraints. """
        if self._root is None:
            self._root = self.get_manager().true()
        return self._root

    def to_formula(self):
        """
        Extracts a LogicFormula from the SDD.
        """
        formula = LogicFormula(keep_order=True)
        root_node = self.get_root_inode()
        i = self._to_formula(formula, root_node, {})
        for n, q, l in self.labeled():
            formula.add_name(n, i, l)

        return formula

    def build_dd(self, root_key=None):
        """
        Build the SDD structure from the current self.nodes starting at index root_key.
        :param root_key: The key of the root node in self.nodes. None if the root must be True.
        """
        # Set root
        if root_key is None:
            root_node = self.get_manager().true()
        else:
            root_node = self.get_inode(root_key)

        self.build_constraint_dd()
        constrained_node = self.get_manager().constraint_dd
        self._root = root_node.conjoin(constrained_node)

    # def cleanup_inodes(self):
    #    """
    #    Dereferences all but the root node and empties self.nodes. ! Do not use self.nodes afterwards !
    #    This is usually called right after the build process, build_dd(root_key).
    #    """
    #    self.get_manager().clean_nodes(self.get_root_inode())

    def to_dot(self, *args, **kwargs):
        if kwargs.get("use_internal"):
            return self.to_internal_dot(node=self.get_root_inode())
        else:
            return self.to_formula().to_dot(*args, **kwargs)

    def sdd_to_dot(self, node, litnamemap=None, show_id=False, merge_leafs=False):
        """
        SDD for the given node, formatted for use with Graphviz dot. This method provides more control over the used
        symbols than to_internal_dot (see litnamemap). Primes are given by a dotted line, subs by a full line.

        :param node: The node to get the dot from.
        :param litnamemap: A dictionary providing the symbols to use. The following options are available:
            1. literals, e.g. {1:'A', -1:'-A', ...},
            2. True/False, e.g. {true':'1', 'false':'0'}
            3. And/Or e.g. {'mult':'x', 'add':'+'}
        :param show_id: Whether to display the ids of each sdd node.
        :param merge_leafs: Whether to merge the same leaf nodes. True results in less nodes but makes it harder to
        render without having crossing lines.
        :return: The dot format of the given node. When node is None, the root node is used.
        :rtype: str
        """
        used_node = node if node is not None else self.get_root_inode()
        return super().sdd_to_dot(
            node=used_node,
            litnamemap=litnamemap,
            show_id=show_id,
            merge_leafs=merge_leafs,
        )


class SDDExplicitManager(SDDManager):
    """
    Manager for SDDs with one root which use the explicit encoding, for example where c :- a,b is represented as
    {c,a,b} v {-c,(-a v -b)}). !Beware!, the self.nodes() of this class might be inaccessible (empty) when calling
    clean_nodes(self, root_inode).
    """

    def __init__(self, varcount=0, auto_gc=False, var_constraint=None):
        """Create a new SDDExplicitManager.

        :param varcount: number of initial variables
        :type varcount: int
        :param auto_gc: use automatic garbage collection and minimization
        :type auto_gc: bool
        :param var_constraint: A variable ordering constraint. Currently only x_constrained namedtuple are allowed.
        :type var_constraint: x_constrained
        """
        SDDManager.__init__(
            self, varcount=varcount, auto_gc=auto_gc, var_constraint=var_constraint
        )

    @staticmethod
    def create_from_SDDManager(sddmgr):
        new_mgr = SDDExplicitManager()
        vars(new_mgr).update(vars(sddmgr))
        return new_mgr

    # Error: wrong refcounts. parent.deref() = all descendents.deref()
    # def clean_nodes(self, root_inode):
    #    """
    #    Cleans up all references except root_inode's. This means all other nodes will be dereferenced.
    #    ! Beware, after calling this self.nodes will be empty, do not use it afterwards !
    #    :param root_inode: The node not to dereference.
    #    :type root_inode: SddNode
    #    """
    #    for node in self.nodes:
    #        if node is not None and node.id != root_inode.id:
    #            while node.ref_count() > 0:
    #                node.deref()
    #    self.nodes = []


class SDDExplicitEvaluator(SDDEvaluator):
    def __init__(self, formula, semiring, weights=None, **kwargs):
        SDDEvaluator.__init__(self, formula, semiring, weights, **kwargs)

    def propagate(self):
        self._initialize()
        self.evaluate_evidence(recompute=True)
        self.normalization = self._evidence_weight

    def _evaluate_evidence(self, recompute=False):
        if self._evidence_weight is None or recompute:
            self._evidence_weight = self._evaluate_root()
            if self.semiring.is_zero(self._evidence_weight):
                raise InconsistentEvidenceError(context=" during compilation")
        return self._evidence_weight

    def evaluate_fact(self, node):
        return self.evaluate(node)

    def evaluate_custom(self, node):
        return self.evaluate(self, node)

    def evaluate(self, node, normalize=True):
        # Trivial case: node is deterministically True or False
        if node == self.formula.TRUE:
            if not self.semiring.is_nsp():
                result = self.semiring.one()
            else:
                # calculates weight of root node.
                result = self._evaluate_root()
                if normalize:
                    result = self.semiring.normalize(result, self._get_z())

        elif node is self.formula.FALSE:
            result = self.semiring.zero()
        else:
            # Set query weight
            index = self.formula.atom2var[abs(node)]
            p_orig, n_orig = self.weights.get(index)
            self._set_value(index, (node > 0))

            # Calculate result
            # print("weights: %s" % self.weights)
            result = self._evaluate_root()

            # Restore query weight
            self.set_weight(index, p_orig, n_orig)

            # print("result: %s" % result)
            # print("normalization: %s" % self._get_z())
            # print("normalizing %s with %s and result before %s" % (normalize, self._get_z(), result))
            if normalize:
                result = self.semiring.normalize(result, self._get_z())
            # print("normalized result: %s" % result)
            # print("----------------------------------------")

        return self.semiring.result(result, self.formula)

    def _evaluate_root(self):
        """
        Evaluate the circuit (root node) with the current weights (self.weights)
        :return: The WMC of the circuit with the current weights
        """
        pr_semiring = isinstance(
            self.semiring, (SemiringLogProbability, SemiringProbability)
        )
        query_def_inode = self.formula.get_root_inode()
        return self._get_manager().wmc(
            query_def_inode,
            self.weights,
            self.semiring,
            pr_semiring=pr_semiring,
            perform_smoothing=True,
            smooth_to_root=True,
        )

    def _reset_value(self, index, pos, neg):
        self.set_weight(index, pos, neg)

    def _set_value(self, index, value):
        """Set value for given node.

        :param index: index of node (for which atom2var[atom] = index)
        :param value: True if we only want a positive weight and set the negative weight to semiring.zero().
                      Else, we only want the negative weight and set the positive weight to semiring.zero().
        """
        if value:
            pos = self.weights[index][0]
            self.set_weight(index, pos, self.semiring.zero())
        else:
            neg = self.weights[index][1]
            self.set_weight(index, self.semiring.zero(), neg)


x_constrained_named = namedtuple(
    "x_constrained", "X_named"
)  # X_named = list of literal names that have to appear before the rest


@transform(LogicDAG, SDDExplicit)
def build_explicit_from_logicdag(source, destination, **kwdargs):
    """Build an SDD2 from a LogicDAG.

    :param source: source formula
    :type source: LogicDAG
    :param destination: destination formula
    :type destination: SDDExplicit
    :param kwdargs: extra arguments
    :return: destination
    :rtype: SDDExplicit
    """

    # Get init varcount
    init_varcount = kwdargs.get("init_varcount", -1)
    var_constraint_named = kwdargs.get("var_constraint", None)

    if var_constraint_named is not None and isinstance(
        var_constraint_named, x_constrained_named
    ):
        var_names = {var: True for var in var_constraint_named.X_named}
        var_ids = []
        atomcount = 1
        for _, clause, c_type in source:
            if clause.name is not None:
                if var_names.get(clause.name, False):
                    var_ids.append(atomcount)
                atomcount += 1
        destination.var_constraint = x_constrained(X=var_ids)

    if init_varcount == -1:
        init_varcount = source.atomcount
        for _, clause, c_type in source:
            if c_type != "atom" and clause.name is not None:
                init_varcount += 1
    destination.init_varcount = init_varcount

    # build
    with Timer("Compiling %s" % destination.__class__.__name__):
        identifier = 0
        line_map = (
            dict()
        )  # line in source mapped to line in destination {src_line: (negated, positive, combined)}
        line = 1  # current line (line_id)
        node_to_indicator = {}  # {node : indicator_node}
        root_nodes = []
        for line_id, clause, c_type in source:
            if c_type == "atom":
                result = destination.add_atom(
                    identifier,
                    clause.probability,
                    clause.group,
                    source.get_name(line_id),
                    cr_extra=False,
                    is_extra=clause.is_extra,
                )
                identifier += 1

                line_map[line_id] = (-result, result, result)
                line += 1
            elif c_type == "conj":
                and_nodes = [
                    line_map[abs(src_line)][src_line > 0]
                    for src_line in clause.children
                ]
                negated_and_nodes = [
                    line_map[abs(src_line)][src_line < 0]
                    for src_line in clause.children
                ]

                if clause.name is None:
                    result = destination.add_and(and_nodes, source.get_name(line_id))
                    result_neg = destination.add_or(
                        negated_and_nodes, source.get_name(-line_id)
                    )
                    line_map[line_id] = (result_neg, result, result)
                    line += 2
                else:
                    # head
                    head = destination.add_atom(
                        identifier=identifier,
                        probability=True,
                        group=None,
                        name=clause.name,
                        is_extra=False,
                    )
                    identifier += 1
                    # body
                    body = destination.add_and(and_nodes)  # source.get_name(i))
                    negated_body = destination.add_or(negated_and_nodes)
                    # combined
                    combined_false = destination.add_and([-head, negated_body])
                    combined_true = destination.add_and([head, body])
                    combined = destination.add_or([combined_false, combined_true])

                    node_to_indicator[combined] = head
                    line_map[line_id] = (combined_false, combined_true, combined)
                    line += 6
                    root_nodes.append(combined)

            elif c_type == "disj":
                or_nodes = [
                    line_map[abs(src_line)][src_line > 0]
                    for src_line in clause.children
                ]
                negated_or_nodes = [
                    line_map[abs(src_line)][src_line < 0]
                    for src_line in clause.children
                ]

                if clause.name is None:
                    result = destination.add_or(or_nodes, source.get_name(line_id))
                    result_neg = destination.add_and(
                        negated_or_nodes, source.get_name(-line_id)
                    )
                    line_map[line_id] = (result_neg, result, result)
                    line += 2
                else:
                    # head
                    head = destination.add_atom(
                        identifier=identifier,
                        probability=True,
                        group=None,
                        name=clause.name,
                    )
                    identifier += 1
                    # body
                    body = destination.add_or(or_nodes)  # source.get_name(i))
                    negated_body = destination.add_and(negated_or_nodes)
                    # combined
                    combined_false = destination.add_and([-head, negated_body])
                    combined_true = destination.add_and([head, body])
                    combined = destination.add_or([combined_false, combined_true])

                    node_to_indicator[combined] = head
                    line_map[line_id] = (combined_false, combined_true, combined)
                    line += 6
                    root_nodes.append(combined)

            else:
                raise TypeError("Unknown node type")

        for name, node, label in source.get_names_with_label():
            if (
                label == destination.LABEL_QUERY
                or label == destination.LABEL_EVIDENCE_MAYBE
                or label == destination.LABEL_EVIDENCE_NEG
                or label == destination.LABEL_EVIDENCE_POS
            ):  # TODO required?
                if node is None or node == 0:
                    destination.add_name(name, node, label)
                else:
                    mapped_line = line_map[abs(node)][2]
                    sign = -1 if node < 0 else 1
                    if (
                        node_to_indicator.get(mapped_line) is not None
                    ):  # Change internal node indicator
                        destination.add_name(
                            name, sign * node_to_indicator[mapped_line], label
                        )
                    else:
                        destination.add_name(name, sign * mapped_line, label)

        if len(root_nodes) > 0:
            root_key = destination.add_and(root_nodes, name=None)
        else:
            root_key = None

        destination.build_dd(root_key)
        # destination.cleanup_inodes()
    return destination


@transform(_ForwardSDD, SDDExplicit)
def build_explicit_from_forwardsdd(source, destination, **kwdargs):
    """Build an SDDExplicit from a _ForwardSDD.

    :param source: source formula
    :type source: _ForwardSDD
    :param destination: destination formula
    :type destination: SDDExplicit
    :param kwdargs: extra arguments
    :return: destination
    :rtype: SDDExplicit
    """
    with Timer("Compiling %s" % destination.__class__.__name__):
        ForwardInference.build_dd(source)

        # Make sure all atoms exist in atom2var.
        for name, node, label in source.labeled():
            if source.is_probabilistic(node):
                source.get_inode(node)

        source.copy_to_noref(destination)
        destination.inode_manager = SDDExplicitManager.create_from_SDDManager(
            destination.inode_manager
        )

        root_nodes = []
        var_to_indicator = dict()
        for line_id, clause, c_type in destination:
            if (c_type == "conj" or c_type == "disj") and clause.name is not None:
                # head
                head = destination.add_atom(
                    identifier=destination.get_next_atom_identifier(),
                    probability=True,
                    group=None,
                    name=destination.get_name(line_id),
                )
                var_to_indicator[line_id] = head

                body = line_id

                combined_true = destination.add_and([head, body])
                combined_false = destination.add_and([-head, -body])
                combined = destination.add_or([combined_false, combined_true])

                root_nodes.append(combined)

        if len(root_nodes) > 0:
            root_key = destination.add_and(root_nodes, name=None)
        else:
            root_key = None

        for name, node, label in source.get_names_with_label():
            new_node = None
            if node is not None:
                new_node = var_to_indicator.get(abs(node), abs(node))
                if node < 0:
                    new_node = -new_node
            destination.add_name(name, new_node, label)

        # query_nodes = destination.queries()
        # evidence_nodes_all = destination.evidence_all()
        # evidence_nodes = destination.evidence()

        destination.build_dd(root_key)
    return destination



"""
Part of the ProbLog distribution.

Copyright 2019 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

from problog import get_evaluatable
from problog.evaluator import SemiringProbability
from problog.formula import LogicFormula
from problog.logic import Term
from problog.program import PrologString

# noinspection PyBroadException
from problog.test.test_system import SemiringProbabilityNSPCopy

try:
    from pysdd import sdd

    has_sdd = True
except Exception as err:
    has_sdd = False

evaluatables = ["ddnnf"]

if has_sdd:
    evaluatables.append("sdd")
    evaluatables.append("sddx")
    evaluatables.append("fsdd")
else:
    print("No SDD support - The evaluator tests are not performed with SDDs.")


class TestEvaluator(unittest.TestCase):
    def test_evaluate_custom_weights(self):
        """
        Tests evaluate() with custom weights (not the ones from the ProbLog file)
        """
        for eval_name in evaluatables:
            with self.subTest(eval_name=eval_name):
                self.evaluate_custom_weights(eval_name)

    def evaluate_custom_weights(self, eval_name=None):
        class TestSemiringProbabilityNSP(SemiringProbability):
            def is_nsp(self):
                return True

            def pos_value(self, a, key=None):
                if isinstance(a, tuple):
                    return float(a[0])
                else:
                    return float(a)

            def neg_value(self, a, key=None):
                if isinstance(a, tuple):
                    return float(a[1])
                else:
                    return 1 - float(a)

        program = """
                    0.25::a.
                    query(a).
                """
        pl = PrologString(program)
        lf = LogicFormula.create_from(pl, label_all=True, avoid_name_clash=True)
        semiring = TestSemiringProbabilityNSP()
        kc_class = get_evaluatable(name=eval_name, semiring=semiring)
        kc = kc_class.create_from(lf)
        a = Term("a")

        # without custom weights
        results = kc.evaluate(semiring=semiring)
        self.assertEqual(0.25, results[a])

        # with custom weights
        weights = {a: 0.1}
        results = kc.evaluate(semiring=semiring, weights=weights)
        self.assertEqual(0.1, results[a])

        # with custom weights
        weights = {a: (0.1, 0.1)}
        results = kc.evaluate(semiring=semiring, weights=weights)
        self.assertEqual(0.5, results[a])

        # with custom weights based on index
        weights = {kc.get_node_by_name(a): 0.2}
        results = kc.evaluate(semiring=semiring, weights=weights)
        self.assertEqual(0.2, results[a])

        # Testing with weight on node 0 (True)
        weights = {0: 0.3, a: (0.1, 0.1)}
        results = kc.evaluate(semiring=semiring, weights=weights)
        self.assertEqual(0.5, results[a])

        # Testing query on node 0 (True)
        class TestSemiringProbabilityIgnoreNormalize(SemiringProbabilityNSPCopy):
            def normalize(self, a, z):
                return a

        weights = {0: (0.3, 0.7), a: (0.1, 0.1)}
        results = kc.evaluate(
            index=0, semiring=TestSemiringProbabilityIgnoreNormalize(), weights=weights
        )
        self.assertEqual(0.06, results)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestEvaluator)
    unittest.TextTestRunner(verbosity=2).run(suite)



import logging
import os
import sqlite3

from problog.errors import UserError, InvalidValue
from problog.extern import problog_export, problog_export_raw
from problog.logic import Term, Constant

logger = logging.getLogger("problog")


def db2pl(dbvalue):
    if type(dbvalue) == str:
        return Term("'" + dbvalue + "'")
    else:
        return Constant(dbvalue)


def pl2db(t):
    if isinstance(t, Constant):
        return t.value
    else:
        return str(t).strip("'")


def get_colnames(conn, tablename):
    cur = conn.cursor()
    cur.execute("SELECT * FROM %s WHERE 0;" % tablename)
    res = [x[0] for x in cur.description]
    cur.close()
    return res


@problog_export("+str")
def sqlite_load(filename):
    filename = problog_export.database.resolve_filename(filename)
    if not os.path.exists(filename):
        raise UserError("Can't find database '%s'" % filename)

    conn = sqlite3.connect(filename)

    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = [x[0] for x in cursor.fetchall()]
    cursor.close()

    for table in tables:
        columns = get_colnames(conn, table)
        types = ["+term"] * len(columns)
        problog_export_raw(*types)(
            QueryFunc(conn, table, columns), funcname=table, modname=None
        )

    return ()


@problog_export("+str", "+str")
def csv_load(filename, predicate):
    import tempfile
    import csv
    import re

    filename = problog_export.database.resolve_filename(filename)
    if not os.path.exists(filename):
        raise UserError("Can't find csv file '%s'" % filename)

    csvfile = open(filename, "r")
    line = 0
    reader = csv.reader(csvfile)
    # Column names
    row = next(reader)
    line += 1
    invalid_chars = re.compile(r"""[^a-zA-Z0-9_]""")
    columns = [invalid_chars.sub("", field) for field in row]

    filepath, ext = os.path.splitext(os.path.basename(filename))
    sql_dir = tempfile.mkdtemp()
    sql_filename = os.path.join(sql_dir, filepath + ".sqlite")
    idx = 1
    while os.path.exists(sql_filename):
        sql_filename = os.path.join(sql_dir, filepath + "_" + str(idx) + ".sqlite")
        idx += 1
    logger.debug("CSV->SQLite: " + sql_filename)
    conn = sqlite3.connect(sql_filename)

    cursor = conn.cursor()

    row = next(reader)
    line += 1
    column_types_sql = []
    column_types_py = []
    for field in row:
        try:
            value = int(field)
            column_types_sql.append("INTEGER")
            column_types_py.append(int)
            continue
        except ValueError:
            pass
        try:
            value = float(field)
            column_types_sql.append("REAL")
            column_types_py.append(float)
            continue
        except ValueError:
            pass
        column_types_sql.append("TEXT")
        column_types_py.append(str)

    coldefs = [n + " " + t for n, t in zip(columns, column_types_sql)]
    cursor.execute("CREATE TABLE " + predicate + "(" + ",".join(coldefs) + ");")

    def insert_statement(row):
        values = []
        for value, sqltype, pytype in zip(row, column_types_sql, column_types_py):
            try:
                if sqltype == "TEXT":
                    values.append("'" + pytype(value.strip()) + "'")
                else:
                    values.append(str(pytype(value)))
            except ValueError:
                raise InvalidValue(
                    "{}:{}. Expected type {}, found value {}".format(
                        filename, line, pytype.__name__, value
                    )
                )
        return (
            "INSERT INTO "
            + predicate
            + "("
            + ",".join(columns)
            + ") VALUES ("
            + ",".join(values)
            + ");"
        )

    cursor.execute(insert_statement(row))
    for row in reader:
        line += 1
        if len(row) == 0:
            continue
        cursor.execute(insert_statement(row))

    conn.commit()
    cursor.close()
    csvfile.close()

    types = ["+term"] * len(columns)
    problog_export_raw(*types)(
        QueryFunc(conn, predicate, columns), funcname=predicate, modname=None
    )

    return ()


class QueryFunc(object):
    def __init__(self, db, tablename, columns):
        self.db = db
        self.tablename = tablename
        self.columns = columns

    def __call__(self, *args, **kwargs):
        where = []
        values = []
        for c, a in zip(self.columns, args):
            if a is not None:
                where.append("%s = ?" % c)
                values.append(pl2db(a))
        where = " AND ".join(where)
        if where:
            where = " WHERE " + where

        query = "SELECT %s FROM %s%s" % (", ".join(self.columns), self.tablename, where)
        cur = self.db.cursor()

        cur.execute(query, values)
        res = [tuple(map(db2pl, r)) for r in cur.fetchall()]
        cur.close()
        return res



import unittest

from problog import get_evaluatable
from problog.program import PrologString


class TestTryCall(unittest.TestCase):
    def test_try_call_existing_fact(self):
        p = PrologString(
            """
        a(1).
        res :- try_call(a(1)).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 1)

    def test_try_call_existing_fact_non_ground(self):
        p = PrologString(
            """
        a(1).
        res :- try_call(a(X)).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 1)

    def test_try_call_existing_fact_non_ground2(self):
        p = PrologString(
            """
        a(1).
        b :- a(X), X > 0.
        res :- try_call(b).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 1)

    def test_try_call_non_existing_fact_non_ground(self):
        p = PrologString(
            """
        a(1).
        b :- a(X), X > 2.
        res :- try_call(b).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 0)

    def test_try_call_non_existing_fact_non_ground2(self):
        p = PrologString(
            """
        a(1).
        res :- try_call(b(X)).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 0)

    def test_try_call_wrong_comp(self):
        p = PrologString(
            """
        res :- try_call(1 > 2).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 0)

    def test_try_call_right_comp(self):
        p = PrologString(
            """
        res :- try_call(1 < 2).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 1)

    def test_try_call_non_existing_comp(self):
        p = PrologString(
            """
        res :- try_call(X > 2).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 0)

    def test_try_call_existing_clause(self):
        p = PrologString(
            """
        a:(c:-d).
        res :- try_call(a:(c:-d)).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 1)

    def test_try_call_non_existing_clause(self):
        p = PrologString(
            """
        a:(c:-d).
        res :- try_call(a:(c:-e)).
        query(res).
        """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values())[0], 0)
